# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Key mathematical operations used for tensorflow backend
======================

Contains the basic linear algebraic utilities needed by the tensorflow simulator backend
Hyperlinks: :class:`Circuit`

.. currentmodule:: strawberryfields.backends.tfbackend.ops

Contents
----------------------
.. autosummary::

"""
# pylint: disable=too-many-arguments

from string import ascii_lowercase as indices
from functools import lru_cache

import tensorflow as tf
import numpy as np
from scipy.special import binom, factorial

from strawberryfields.backends.shared_ops import (
    generate_bs_factors,
    load_bs_factors,
    save_bs_factors,
    squeeze_parity,
)

def_type = tf.complex64
max_num_indices = len(indices)

###################################################################

# Helper functions:


def _numer_safe_power(base, exponent):
    """gives the desired behaviour of 0**0=1"""
    if exponent == 0:
        return tf.ones_like(base, dtype=def_type)

    return base ** exponent


def mixed(pure_state, batched=False):
    """Converts the state from pure to mixed"""
    # todo: In the fock backend mixing is done by ops.mix(), maybe the functions should be named identically?
    if not batched:
        pure_state = tf.expand_dims(pure_state, 0)  # add in fake batch dimension
    batch_offset = 1
    num_modes = len(pure_state.shape) - batch_offset
    max_num = (max_num_indices - batch_offset) // 2
    if num_modes > max_num:
        raise ValueError(
            "Converting state from pure to mixed currently only supported for {} modes.".format(
                max_num
            )
        )
    else:
        # eqn: 'abc...xyz,ABC...XYZ->aAbBcC...xXyYzZ' (lowercase belonging to 'bra' side, uppercase belonging to 'ket' side)
        batch_index = indices[:batch_offset]
        bra_indices = indices[batch_offset : batch_offset + num_modes]
        ket_indices = indices[batch_offset + num_modes : batch_offset + 2 * num_modes]
        eqn_lhs = batch_index + bra_indices + "," + batch_index + ket_indices
        eqn_rhs = "".join(bdx + kdx for bdx, kdx in zip(bra_indices, ket_indices))
        eqn = eqn_lhs + "->" + batch_index + eqn_rhs
        mixed_state = tf.einsum(eqn, pure_state, tf.conj(pure_state))
    if not batched:
        mixed_state = tf.squeeze(mixed_state, 0)  # drop fake batch dimension
    return mixed_state


def batchify_indices(idxs, batch_size):
    """adds batch indices to the index numbering"""
    return [(bdx,) + idxs[i] for bdx in range(batch_size) for i in range(len(idxs))]


def unravel_index(ind, tensor_shape):
    """no official tensorflow implementation for this yet;
    this is based on one proposed in https://github.com/tensorflow/tensorflow/issues/2075
    """
    ind = tf.expand_dims(tf.cast(ind, tf.int64), 0)
    tensor_shape = tf.expand_dims(tf.cast(tensor_shape, tf.int64), 1)
    strides = tf.cumprod(tensor_shape, reverse=True)
    strides_shifted = tf.cumprod(tensor_shape, exclusive=True, reverse=True)
    unraveled_coords = (ind % strides) // strides_shifted
    return tf.transpose(unraveled_coords)


@lru_cache()
def get_prefac_tensor(D, directory, save):
    """Equivalent to the functionality of shared_ops the bs_factors functions from shared_ops,
    but caches the return value as a tensor. This allows us to re-use the same prefactors and save
    space on the computational graph."""
    try:
        prefac = load_bs_factors(D, directory)
    except FileNotFoundError:
        prefac = generate_bs_factors(D)
        if save:
            save_bs_factors(prefac, directory)
    prefac = tf.expand_dims(tf.cast(prefac[:D, :D, :D, :D, :D], def_type), 0)
    return prefac


###################################################################

# Matrices:
# These return a matrix of shape [D, D]
# which transforms from the input (truncated) Fock basis to the
# output (truncated) Fock basis.


def squeezed_vacuum_vector(r, theta, D, batched=False, eps=1e-32):
    """returns the ket representing a single mode squeezed vacuum state"""
    if batched:
        batch_size = r.shape[0].value
    r = tf.cast(r, def_type)
    theta = tf.cast(theta, def_type)
    c1 = tf.cast(
        tf.stack(
            [
                tf.sqrt(1 / tf.cosh(r)) * np.sqrt(factorial(k)) / factorial(k / 2.0)
                for k in range(0, D, 2)
            ],
            axis=-1,
        ),
        def_type,
    )
    c2 = tf.stack(
        [
            (-0.5 * tf.exp(1j * theta) * tf.cast(tf.tanh(r + eps), def_type)) ** (k / 2.0)
            for k in range(0, D, 2)
        ],
        axis=-1,
    )
    even_coeffs = c1 * c2
    ind = [(k,) for k in np.arange(0, D, 2)]
    shape = [D]
    if batched:
        ind = batchify_indices(ind, batch_size)
        shape = [batch_size] + shape
    output = tf.scatter_nd(ind, tf.reshape(even_coeffs, [-1]), shape)
    return output


def squeezer_matrix(r, theta, D, batched=False):
    """creates the single mode squeeze matrix"""
    r = tf.cast(r, tf.float64)
    if not batched:
        r = tf.expand_dims(r, 0)  # introduce artificial batch dimension
    r = tf.reshape(r, [-1, 1, 1, 1])
    theta = tf.cast(theta, def_type)
    theta = tf.reshape(theta, [-1, 1, 1, 1])

    rng = np.arange(D)
    n = np.reshape(rng, [-1, D, 1, 1])
    m = np.reshape(rng, [-1, 1, D, 1])
    k = np.reshape(rng, [-1, 1, 1, D])

    phase = tf.exp(1j * theta * (n - m) / 2)
    signs = squeeze_parity(D).reshape([1, D, 1, D])
    mask = np.logical_and(
        (m + n) % 2 == 0, k <= np.minimum(m, n)
    )  # kills off terms where the sum index k goes past min(m,n)
    k_terms = (
        signs
        * tf.pow(tf.sinh(r) / 2, mask * (n + m - 2 * k) / 2)
        * mask
        / tf.pow(tf.cosh(r), (n + m + 1) / 2)
        * tf.exp(
            0.5 * tf.lgamma(tf.cast(m + 1, tf.float64))
            + 0.5 * tf.lgamma(tf.cast(n + 1, tf.float64))
            - tf.lgamma(tf.cast(k + 1, tf.float64))
            - tf.lgamma(tf.cast((m - k) / 2 + 1, tf.float64))
            - tf.lgamma(tf.cast((n - k) / 2 + 1, tf.float64))
        )
    )
    output = tf.reduce_sum(phase * tf.cast(k_terms, def_type), axis=-1)

    if not batched:
        # remove extra batch dimension
        output = tf.squeeze(output, 0)
    return output


def phase_shifter_matrix(theta, D, batched=False):
    """creates the single mode phase shifter matrix"""
    if batched:
        batch_size = theta.shape[0].value
    theta = tf.cast(theta, def_type)
    shape = [D, D]
    if batched:
        shape = [batch_size] + shape
    zero_matrix = tf.zeros(shape=shape, dtype=def_type)
    diag = [tf.exp(1j * theta * k) for k in np.arange(D, dtype=np.complex64)]
    if batched:
        diag = tf.stack(diag, axis=1)
    diag_matrix = tf.matrix_set_diag(zero_matrix, diag)
    return diag_matrix


def kerr_interaction_matrix(kappa, D, batched=False):
    """creates the single mode Kerr interaction matrix"""
    coeffs = [tf.exp(1j * kappa * n ** 2) for n in range(D)]
    if batched:
        coeffs = tf.stack(coeffs, axis=1)
    output = tf.matrix_diag(coeffs)
    return output


def cross_kerr_interaction_matrix(kappa, D, batched=False):
    """creates the two mode cross-Kerr interaction matrix"""
    coeffs = [tf.exp(1j * kappa * n1 * n2) for n1 in range(D) for n2 in range(D)]
    if batched:
        coeffs = tf.stack(coeffs, axis=1)
    output = tf.matrix_diag(coeffs)
    if batched:
        output = tf.transpose(tf.reshape(output, [-1] + [D] * 4), [0, 1, 3, 2, 4])
    else:
        output = tf.transpose(tf.reshape(output, [D] * 4), [0, 2, 1, 3])
    return output


def cubic_phase_matrix(gamma, D, hbar, batched=False, method="self_adjoint_eig"):
    """creates the single mode cubic phase matrix"""
    a, ad = ladder_ops(D)
    x = np.sqrt(hbar / 2) * tf.cast(a + ad, def_type)
    x3 = x @ x @ x
    if batched:
        x3 = tf.expand_dims(x3, 0)
        gamma = tf.reshape(gamma, [-1, 1, 1])
    H0 = gamma / (3 * hbar) * x3
    lambdas, U = tf.self_adjoint_eig(H0)
    transpose_list = [1, 0]
    if batched:
        transpose_list = [0, 2, 1]
    if method == "self_adjoint_eig":
        # This seems to work as long as the matrix has dimension 18x18 or smaller
        # For larger matrices, Tensorflow returns the error: 'Self-adjoint eigen decomposition was not successful.'
        V = U @ tf.matrix_diag(tf.exp(1j * lambdas)) @ tf.conj(tf.transpose(U, transpose_list))
    # below version works for matrices larger than 18x18, but
    # expm was not added until TF v1.5, while
    # as of TF v1.5, gradient of expm is not implemented
    # elif method == "expm":
    #    V = tf.linalg.expm(1j * H0)
    else:
        raise ValueError("'method' must be either 'self_adjoint_eig' or 'expm'.")
    return V


def loss_superop(T, D, batched=False):
    """creates the single mode loss channel matrix"""
    # indices are abcd, corresponding to action |a><b| (.) |c><d|
    if not batched:
        T = tf.expand_dims(T, 0)  # introduce artificial batch dimension
    T = tf.reshape(T, [-1, 1, 1, 1, 1, 1])
    T = tf.cast(T, tf.float64)
    rng = np.arange(D)
    a = np.reshape(rng, [-1, D, 1, 1, 1, 1])
    b = np.reshape(rng, [-1, 1, D, 1, 1, 1])
    c = np.reshape(rng, [-1, 1, 1, D, 1, 1])
    d = np.reshape(rng, [-1, 1, 1, 1, D, 1])
    l = np.reshape(rng, [-1, 1, 1, 1, 1, D])
    a_mask = np.where(a == b - l, 1, 0)
    d_mask = np.where(d == c - l, 1, 0)
    mask = (l <= np.minimum(b, c)) * a_mask * d_mask

    exponent = tf.abs((b + c) / 2 - l) * mask
    T_numer = tf.pow((1 - T), l) * tf.pow(T, exponent)
    fact_numer = np.sqrt(factorial(b) * factorial(c))
    fact_denom = np.sqrt(factorial(b - l) * factorial(c - l)) * factorial(l)
    fact_denom_masked = np.where(fact_denom > 0, fact_denom, 1)
    factors = mask * fact_numer / fact_denom_masked
    l_terms = T_numer * factors
    output = tf.cast(tf.reduce_sum(l_terms, -1), def_type)
    if not batched:
        output = tf.squeeze(output, 0)  # drop artificial batch dimension
    return output


def displacement_matrix(alpha, D, batched=False):
    """creates the single mode displacement matrix"""
    if batched:
        batch_size = alpha.shape[0].value
    alpha = tf.cast(alpha, def_type)
    idxs = [(j, k) for j in range(D) for k in range(j)]
    values = [
        alpha ** (j - k) * tf.cast(tf.sqrt(binom(j, k) / factorial(j - k)), def_type)
        for j in range(D)
        for k in range(j)
    ]
    values = tf.stack(values, axis=-1)
    dense_shape = [D, D]
    vals = [1.0] * D
    ind = [(idx, idx) for idx in range(D)]
    if batched:
        dense_shape = [batch_size] + dense_shape
        vals = vals * batch_size
        ind = batchify_indices(ind, batch_size)
    eye_diag = tf.SparseTensor(ind, vals, dense_shape)
    signs = [(-1) ** (j - k) for j in range(D) for k in range(j)]
    if batched:
        idxs = batchify_indices(idxs, batch_size)
        signs = signs * batch_size
    sign_lower_diag = tf.cast(tf.SparseTensor(idxs, signs, dense_shape), tf.float32)
    sign_matrix = tf.sparse_add(eye_diag, sign_lower_diag)
    sign_matrix = tf.cast(tf.sparse_tensor_to_dense(sign_matrix), def_type)
    lower_diag = tf.scatter_nd(idxs, tf.reshape(values, [-1]), dense_shape)
    E = tf.cast(tf.eye(D), def_type) + lower_diag
    E_prime = tf.conj(E) * sign_matrix
    if batched:
        eqn = "aik,ajk->aij"  # pylint: disable=bad-whitespace
    else:
        eqn = "ik,jk->ij"  # pylint: disable=bad-whitespace
    prefactor = tf.expand_dims(
        tf.expand_dims(tf.cast(tf.exp(-0.5 * tf.abs(alpha) ** 2), def_type), -1), -1
    )
    D_alpha = prefactor * tf.einsum(eqn, E, E_prime)
    return D_alpha


def beamsplitter_matrix(t, r, D, batched=False, save=False, directory=None):
    """creates the two mode beamsplitter matrix"""
    if not batched:
        # put in a fake batch dimension for broadcasting convenience
        t = tf.expand_dims(t, 0)
        r = tf.expand_dims(r, 0)
    t = tf.cast(tf.reshape(t, [-1, 1, 1, 1, 1, 1]), def_type)
    r = tf.cast(tf.reshape(r, [-1, 1, 1, 1, 1, 1]), def_type)
    mag_t = tf.cast(t, tf.float32)
    mag_r = tf.abs(r)
    phase_r = tf.atan2(tf.imag(r), tf.real(r))

    rng = tf.range(D, dtype=tf.float32)
    N = tf.reshape(rng, [1, -1, 1, 1, 1, 1])
    n = tf.reshape(rng, [1, 1, -1, 1, 1, 1])
    M = tf.reshape(rng, [1, 1, 1, -1, 1, 1])
    k = tf.reshape(rng, [1, 1, 1, 1, 1, -1])
    n_minus_k = n - k
    N_minus_k = N - k
    M_minus_n_plus_k = M - n + k
    # need to deal with 0*(-n) for n integer
    n_minus_k = tf.where(tf.greater(n, k), n_minus_k, tf.zeros_like(n_minus_k))
    N_minus_k = tf.where(tf.greater(N, k), N_minus_k, tf.zeros_like(N_minus_k))
    M_minus_n_plus_k = tf.where(
        tf.greater(M_minus_n_plus_k, 0), M_minus_n_plus_k, tf.zeros_like(M_minus_n_plus_k)
    )

    powers = tf.cast(
        tf.pow(mag_t, k)
        * tf.pow(mag_r, n_minus_k)
        * tf.pow(mag_r, N_minus_k)
        * tf.pow(mag_t, M_minus_n_plus_k),
        def_type,
    )
    phase = tf.exp(1j * tf.cast(phase_r * (n - N), def_type))

    # load parameter-independent prefactors
    prefac = get_prefac_tensor(D, directory, save)

    if prefac.graph != phase.graph:
        # if cached prefactors live on another graph, we'll have to reload them into this graph.
        # In future versions, if 'copy_variable_to_graph' comes out of contrib, consider using that
        get_prefac_tensor.cache_clear()
        prefac = get_prefac_tensor(D, directory, save)

    BS_matrix = tf.reduce_sum(phase * powers * prefac, -1)

    if not batched:
        # drop artificial batch index
        BS_matrix = tf.squeeze(BS_matrix, [0])
    return BS_matrix


###################################################################

# Input states:
# Creates input states on a single mode


def fock_state(n, D, pure=True, batched=False):
    """creates a single mode input Fock state"""
    if not isinstance(n, (np.ndarray, int)):
        raise ValueError("'n' is expected to be either an int or a numpy array")
    if batched:
        batch_size = n.shape[0]
        idxs = [(b, f) for (b, f) in zip(range(batch_size), n)]
        values = [1.0] * batch_size
        shape = [batch_size, D]
    else:
        idxs = [(n,)]
        values = [1.0]
        shape = [D]
    fock_sparse = tf.scatter_nd(idxs, values, shape)
    fock = tf.cast(fock_sparse, def_type)
    if not pure:
        fock = mixed(fock, batched)
    return fock


def coherent_state(alpha, D, pure=True, batched=False):
    """creates a single mode input coherent state"""
    coh = tf.stack(
        [
            tf.exp(-0.5 * tf.conj(alpha) * alpha)
            * _numer_safe_power(alpha, n)
            / tf.cast(np.sqrt(factorial(n)), def_type)
            for n in range(D)
        ],
        axis=-1,
    )
    if not pure:
        coh = mixed(coh, batched)
    return coh


def squeezed_vacuum(r, theta, D, pure=True, batched=False):
    """creates a single mode input squeezed vacuum state"""
    squeezed = squeezed_vacuum_vector(r, theta, D, batched=batched)
    if not pure:
        squeezed = mixed(squeezed, batched)
    return squeezed


def displaced_squeezed(alpha, r, phi, D, pure=True, batched=False, eps=1e-12):
    """creates a single mode input displaced squeezed state"""
    alpha = tf.cast(alpha, def_type)
    r = (
        tf.cast(r, def_type) + eps
    )  # to prevent nans if r==0, we add an epsilon (default is miniscule)
    phi = tf.cast(phi, def_type)

    phase = tf.exp(1j * phi)
    sinh = tf.sinh(r)
    cosh = tf.cosh(r)
    tanh = tf.tanh(r)

    # create Hermite polynomials
    gamma = alpha * cosh + tf.conj(alpha) * phase * sinh
    hermite_arg = gamma / tf.sqrt(phase * tf.sinh(2 * r))

    prefactor = tf.expand_dims(
        tf.exp(-0.5 * alpha * tf.conj(alpha) - 0.5 * tf.conj(alpha) ** 2 * phase * tanh), -1
    )
    coeff = tf.stack(
        [
            _numer_safe_power(0.5 * phase * tanh, n / 2.0) / tf.sqrt(factorial(n) * cosh)
            for n in range(D)
        ],
        axis=-1,
    )
    hermite_terms = tf.stack([tf.cast(H(n, hermite_arg), def_type) for n in range(D)], axis=-1)
    squeezed_coh = prefactor * coeff * hermite_terms

    if not pure:
        squeezed_coh = mixed(squeezed_coh, batched)
    return squeezed_coh


def thermal_state(nbar, D):
    """creates a single mode input thermal state.
    Note that the batch dimension is determined by argument nbar.
    """
    nbar = tf.cast(nbar, def_type)
    coeffs = tf.stack(
        [_numer_safe_power(nbar, n) / _numer_safe_power(nbar + 1, n + 1) for n in range(D)], axis=-1
    )
    thermal = tf.matrix_diag(coeffs)
    return thermal


###################################################################

# Generic Gate implementations:
# These apply the given matrix to the specified mode(s) of in_modes


def single_mode_gate(matrix, mode, in_modes, pure=True, batched=False):
    """basic form:
    'ab,cde...b...xyz->cde...a...xyz' (pure state)
    'ab,ef...bc...xyz,cd->ef...ad...xyz' (mixed state)
    """
    if batched:
        batch_offset = 1
    else:
        batch_offset = 0
    batch_index = indices[:batch_offset]
    left_gate_str = indices[batch_offset : batch_offset + 2]  # |a><b|
    num_indices = len(in_modes.shape)
    if pure:
        num_modes = num_indices - batch_offset
        mode_size = 1
    else:
        right_gate_str = indices[batch_offset + 2 : batch_offset + 4]  # |c><d|
        num_modes = (num_indices - batch_offset) // 2
        mode_size = 2
    max_len = len(indices) - 2 * mode_size - batch_offset
    if num_modes == 0:
        raise ValueError("'in_modes' must have at least one mode")
    if num_modes > max_len:
        raise NotImplementedError(
            "The max number of supported modes for this operation is currently {}".format(max_len)
        )
    if mode < 0 or mode >= num_modes:
        raise ValueError("'mode' argument is not compatible with number of in_modes")
    else:
        other_modes_indices = indices[
            batch_offset + 2 * mode_size : batch_offset + (1 + num_modes) * mode_size
        ]
        if pure:
            eqn_lhs = "{},{}{}{}{}".format(
                batch_index + left_gate_str,
                batch_index,
                other_modes_indices[: mode * mode_size],
                left_gate_str[1],
                other_modes_indices[mode * mode_size :],
            )
            eqn_rhs = "".join(
                [
                    batch_index,
                    other_modes_indices[: mode * mode_size],
                    left_gate_str[0],
                    other_modes_indices[mode * mode_size :],
                ]
            )
        else:
            eqn_lhs = "{},{}{}{}{}{},{}".format(
                batch_index + left_gate_str,
                batch_index,
                other_modes_indices[: mode * mode_size],
                left_gate_str[1],
                right_gate_str[0],
                other_modes_indices[mode * mode_size :],
                batch_index + right_gate_str,
            )
            eqn_rhs = "".join(
                [
                    batch_index,
                    other_modes_indices[: mode * mode_size],
                    left_gate_str[0],
                    right_gate_str[1],
                    other_modes_indices[mode * mode_size :],
                ]
            )

    eqn = eqn_lhs + "->" + eqn_rhs
    einsum_inputs = [matrix, in_modes]
    if not pure:
        transposed_axis = [0, 2, 1] if batched else [1, 0]
        einsum_inputs.append(tf.transpose(tf.conj(matrix), transposed_axis))
    output = tf.einsum(eqn, *einsum_inputs)
    return output


def two_mode_gate(matrix, mode1, mode2, in_modes, pure=True, batched=False):
    """basic form:
    'abcd,efg...b...d...xyz->efg...a...c...xyz' (pure state)
    'abcd,ij...be...dg...xyz,efgh->ij...af...ch...xyz' (mixed state)
    """
    # pylint: disable=too-many-branches,too-many-statements
    if batched:
        batch_offset = 1
    else:
        batch_offset = 0
    batch_index = indices[:batch_offset]
    left_gate_str = indices[batch_offset : batch_offset + 4]  # |a><b| |c><d|
    num_indices = len(in_modes.shape)
    if pure:
        num_modes = num_indices - batch_offset
        mode_size = 1
    else:
        right_gate_str = indices[batch_offset + 4 : batch_offset + 8]  # |e><f| |g><h|
        num_modes = (num_indices - batch_offset) // 2
        mode_size = 2
    max_len = (len(indices) - 4) // mode_size - batch_offset

    if num_modes == 0:
        raise ValueError("'in_modes' must have at least one mode")
    if num_modes > max_len:
        raise NotImplementedError(
            "The max number of supported modes for this operation is currently {}".format(max_len)
        )
    else:
        min_mode = min(mode1, mode2)
        max_mode = max(mode1, mode2)
        if min_mode < 0 or max_mode >= num_modes or mode1 == mode2:
            raise ValueError("One or more mode numbers are incompatible")
        else:
            other_modes_indices = indices[
                batch_offset
                + 4 * mode_size : batch_offset
                + 4 * mode_size
                + mode_size * (num_modes - 2)
            ]
            # build equation
            if mode1 == min_mode:
                lhs_min_mode_indices = left_gate_str[1]
                lhs_max_mode_indices = left_gate_str[3]
                rhs_min_mode_indices = left_gate_str[0]
                rhs_max_mode_indices = left_gate_str[2]
            else:
                lhs_min_mode_indices = left_gate_str[3]
                lhs_max_mode_indices = left_gate_str[1]
                rhs_min_mode_indices = left_gate_str[2]
                rhs_max_mode_indices = left_gate_str[0]
            if not pure:
                if mode1 == min_mode:
                    lhs_min_mode_indices += right_gate_str[0]
                    lhs_max_mode_indices += right_gate_str[2]
                    rhs_min_mode_indices += right_gate_str[1]
                    rhs_max_mode_indices += right_gate_str[3]
                else:
                    lhs_min_mode_indices += right_gate_str[2]
                    lhs_max_mode_indices += right_gate_str[0]
                    rhs_min_mode_indices += right_gate_str[3]
                    rhs_max_mode_indices += right_gate_str[1]
            eqn_lhs = "{},{}{}{}{}{}{}".format(
                batch_index + left_gate_str,
                batch_index,
                other_modes_indices[: min_mode * mode_size],
                lhs_min_mode_indices,
                other_modes_indices[min_mode * mode_size : (max_mode - 1) * mode_size],
                lhs_max_mode_indices,
                other_modes_indices[(max_mode - 1) * mode_size :],
            )
            if not pure:
                eqn_lhs += "," + batch_index + right_gate_str
            eqn_rhs = "".join(
                [
                    batch_index,
                    other_modes_indices[: min_mode * mode_size],
                    rhs_min_mode_indices,
                    other_modes_indices[min_mode * mode_size : (max_mode - 1) * mode_size],
                    rhs_max_mode_indices,
                    other_modes_indices[(max_mode - 1) * mode_size :],
                ]
            )
            eqn = eqn_lhs + "->" + eqn_rhs
            einsum_inputs = [matrix, in_modes]
            if not pure:
                if batched:
                    transpose_list = [0, 2, 1, 4, 3]
                else:
                    transpose_list = [1, 0, 3, 2]
                einsum_inputs.append(tf.conj(tf.transpose(matrix, transpose_list)))
            output = tf.einsum(eqn, *einsum_inputs)
            return output


def single_mode_superop(superop, mode, in_modes, pure=True, batched=False):
    """rho_out = S[rho_in]
    (state is always converted to mixed to apply superop)
    basic form:
    abcd,ef...klbcmn...yz->ef...kladmn...yz
    """

    if batched:
        batch_offset = 1
    else:
        batch_offset = 0

    max_len = (len(indices) - 2) // 2 - batch_offset
    if pure:
        num_modes = len(in_modes.shape) - batch_offset
    else:
        num_modes = (len(in_modes.shape) - batch_offset) // 2
    if num_modes > max_len:
        raise NotImplementedError(
            "The max number of supported modes for this operation is currently {}".format(max_len)
        )
    else:
        if pure:
            in_modes = mixed(in_modes, batched)

        # create equation
        batch_index = indices[:batch_offset]
        superop_indices = indices[batch_offset : batch_offset + 4]
        state_indices = indices[batch_offset + 4 : batch_offset + 4 + 2 * num_modes]
        left_unchanged_indices = state_indices[: 2 * mode]
        right_unchanged_indices = state_indices[2 * mode : 2 * (num_modes - 1)]
        eqn_lhs = ",".join(
            [
                batch_index + superop_indices,
                batch_index
                + left_unchanged_indices
                + superop_indices[1:3]
                + right_unchanged_indices,
            ]
        )
        eqn_rhs = "".join(
            [
                batch_index,
                left_unchanged_indices
                + superop_indices[0]
                + superop_indices[3]
                + right_unchanged_indices,
            ]
        )
        eqn = "->".join([eqn_lhs, eqn_rhs])
        new_state = tf.einsum(eqn, superop, in_modes)
        return new_state


###################################################################

# Quantum Gate/Channel functions:
# Primary quantum optical gates implemented as transformations on 'in_modes'


def phase_shifter(theta, mode, in_modes, D, pure=True, batched=False):
    """returns phase shift unitary matrix on specified input modes"""
    matrix = phase_shifter_matrix(theta, D, batched=batched)
    output = single_mode_gate(matrix, mode, in_modes, pure, batched)
    return output


def displacement(alpha, mode, in_modes, D, pure=True, batched=False):
    """returns displacement unitary matrix on specified input modes"""
    matrix = displacement_matrix(alpha, D, batched)
    output = single_mode_gate(matrix, mode, in_modes, pure, batched)
    return output


def squeezer(r, theta, mode, in_modes, D, pure=True, batched=False):
    """returns squeezer unitary matrix on specified input modes"""
    matrix = squeezer_matrix(r, theta, D, batched)
    output = single_mode_gate(matrix, mode, in_modes, pure, batched)
    return output


def kerr_interaction(kappa, mode, in_modes, D, pure=True, batched=False):
    """returns Kerr unitary matrix on specified input modes"""
    matrix = kerr_interaction_matrix(kappa, D, batched)
    output = single_mode_gate(matrix, mode, in_modes, pure, batched)
    return output


def cross_kerr_interaction(kappa, mode1, mode2, in_modes, D, pure=True, batched=False):
    """returns cross-Kerr unitary matrix on specified input modes"""
    matrix = cross_kerr_interaction_matrix(kappa, D, batched)
    output = two_mode_gate(matrix, mode1, mode2, in_modes, pure, batched)
    return output


def cubic_phase(
    gamma, mode, in_modes, D, hbar=2, pure=True, batched=False, method="self_adjoint_eig"
):
    """returns cubic phase unitary matrix on specified input modes"""
    matrix = cubic_phase_matrix(gamma, D, hbar, batched, method=method)
    output = single_mode_gate(matrix, mode, in_modes, pure, batched)
    return output


def beamsplitter(t, r, mode1, mode2, in_modes, D, pure=True, batched=False):
    """returns beamsplitter unitary matrix on specified input modes"""
    matrix = beamsplitter_matrix(t, r, D, batched)
    output = two_mode_gate(matrix, mode1, mode2, in_modes, pure, batched)
    return output


def loss_channel(T, mode, in_modes, D, pure=True, batched=False):
    """returns loss channel matrix on specified input modes"""
    superop = loss_superop(T, D, batched)
    output = single_mode_superop(superop, mode, in_modes, pure, batched)
    return output


###################################################################

# Other useful functions for organizing modes


def combine_single_modes(modes_list, batched=False):
    """Group together a list of single modes (each having dim=1 or dim=2) into a composite mode system."""
    if batched:
        batch_offset = 1
    else:
        batch_offset = 0
    num_modes = len(modes_list)
    if num_modes <= 1:
        raise ValueError("'modes_list' must have at least two modes")

    dims = np.array([len(mode.shape) - batch_offset for mode in modes_list])
    if min(dims) < 1 or max(dims) > 2:
        raise ValueError("Each mode in 'modes_list' can only have dim=1 or dim=2")

    if np.all(dims == 1):
        # All modes are represented as pure states.
        # Can return combined state also as pure state.
        # basic form:
        # 'a,b,c,...,x,y,z->abc...xyz'
        max_num = max_num_indices - batch_offset
        if num_modes > max_num:
            raise NotImplementedError(
                "The max number of supported modes for this operation with pure states is currently {}".format(
                    max_num
                )
            )
        batch_index = indices[:batch_offset]
        out_str = indices[batch_offset : batch_offset + num_modes]
        modes_str = ",".join([batch_index + idx for idx in out_str])
        eqn = "{}->{}".format(modes_str, batch_index + out_str)
        einsum_inputs = modes_list
    else:
        # Some modes are mixed.
        # Return combined state as mixed.
        # basic form:
        # e.g., if first mode is pure and second is mixed...
        # 'a,b,cd,...->abcd...'
        # where (a,b) will belong to the first mode (bra & ket)
        # and cd will belong to the second mode (density matrix)
        max_num = (max_num_indices - batch_offset) // 2
        batch_index = indices[:batch_offset]
        if num_modes > max_num:
            raise NotImplementedError(
                "The max number of supported modes for this operation with mixed states is currently {}".format(
                    max_num
                )
            )
        mode_idxs = [
            indices[slice(batch_offset + idx, batch_offset + idx + 2)]
            for idx in range(0, 2 * num_modes, 2)
        ]  # each mode gets a pair of consecutive indices
        eqn_rhs = batch_index + "".join(mode_idxs)
        eqn_idxs = [
            batch_index + m if dims[idx] == 2 else ",".join(m) for idx, m in enumerate(mode_idxs)
        ]
        eqn_lhs = ",".join(eqn_idxs)
        eqn = eqn_lhs + "->" + eqn_rhs
        einsum_inputs = []
        for idx, mode in enumerate(modes_list):
            if dims[idx] == 1:
                new_inputs = [mode, tf.conj(mode)]
            elif dims[idx] == 2:
                new_inputs = [mode]
            einsum_inputs += new_inputs
    combined_modes = tf.einsum(eqn, *einsum_inputs)
    return combined_modes


def replace_mode(replacement, mode, system, state_is_pure, batched=False):
    """Replace the subsystem 'mode' of 'system' with new state 'replacement'. Argument 'state_is_pure' indicates whether

    This is just a simple wrapper for replace_modes()
    """
    # deprecated: This method has become obsolete and is superseeded by replace_modes()
    if isinstance(mode, int):
        mode = [mode]
    replace_modes(replacement, mode, system, state_is_pure, batched)


def replace_modes(replacement, modes, system, system_is_pure, batched=False):
    """Replace the subsystem 'mode' of 'system' with new state 'replacement'. Argument 'system_is_pure' indicates whether
    'system' is represented by a pure state or a density matrix.
    Note: Does not check if this replacement is physically valid (i.e., if 'replacement' is a valid state)
    Note: expects the shape of both replacement and system to match the batched parameter
    Note: modes does not need to be ordered.
    """
    # todo: write a better docstring
    if isinstance(modes, int):
        modes = [modes]

    if batched:
        batch_offset = 1
    else:
        batch_offset = 0

    num_modes = len(system.shape) - batch_offset
    if not system_is_pure:
        num_modes = int(num_modes / 2)

    replacement_is_pure = bool(len(replacement.shape) - batch_offset == len(modes))

    # we take care of sorting modes below
    if len(modes) == num_modes:
        # we are replacing all the modes
        revised_modes = replacement
        revised_modes_pure = replacement_is_pure
    else:
        # we are replacing a subset, so have to trace
        # make both system and replacement mixed
        # todo: For performance the partial trace could be done directly from the pure state. This would of course require a better partial trace function...
        if system_is_pure:
            system = mixed(system, batched)

        # mix the replacement if it is pure
        if replacement_is_pure:
            replacement = mixed(replacement, batched)

        # partial trace out modes
        # todo: We are tracing out the modes one by one in descending order (to not screw up things). This is quite inefficient.
        reduced_state = system
        for mode in sorted(modes, reverse=True):
            reduced_state = partial_trace(reduced_state, mode, False, batched)
        # append mode and insert state (There is also insert_state(), but it seemed unnecessarily complicated to try to generalize this function, which does a lot of manual list comprehension, to the multi mode case than to write the two liner below)
        # todo: insert_state() could be rewritten to take full advantage of the high level functions of tf. Before doing that different implementatinos should be benchmarked to compare speed and memory requirements, as in practice these methods will be perfomance critical.
        if not batched:
            # todo: remove the hack in the line below and enable the line with axes=0 instead, if ever we change the dependency of SF to tensorflow>=1.6
            # revised_modes = tf.tensordot(reduced_state, replacement, axes=0)
            revised_modes = tf.tensordot(
                tf.expand_dims(reduced_state, 0), tf.expand_dims(replacement, 0), axes=[[0], [0]]
            )
        else:
            batch_size = reduced_state.shape[0].value
            # todo: remove the hack in the line below and enabled the line with axes=0 instead, if ever we change the dependency of SF to tensorflow>=1.6
            # revised_modes = tf.stack([tf.tensordot(reduced_state[b], replacement[b], axes=0) for b in range(batch_size)])
            revised_modes = tf.stack(
                [
                    tf.tensordot(
                        tf.expand_dims(reduced_state[b], 0),
                        tf.expand_dims(replacement[b], 0),
                        axes=[[0], [0]],
                    )
                    for b in range(batch_size)
                ]
            )
        revised_modes_pure = False

    # unless the preparation was meant to go into the last modes in the standard order, we need to swap indices around
    if modes != list(range(num_modes - len(modes), num_modes)):
        mode_permutation = [x for x in range(num_modes) if x not in modes] + modes
        revised_modes = reorder_modes(revised_modes, mode_permutation, revised_modes_pure, batched)

    return revised_modes


def reorder_modes(state, mode_permutation, pure, batched):
    """Reorder the indices of states according to the given mode_permutation, needs to know whether the state is pure and/or batched."""
    if pure:
        scale = 1
        index_permutation = mode_permutation
    else:
        scale = 2
        index_permutation = [
            scale * x + i for x in mode_permutation for i in (0, 1)
        ]  # two indices per mode if we have pure states

    if batched:
        index_permutation = [0] + [i + 1 for i in index_permutation]

    index_permutation = np.argsort(index_permutation)

    reordered_modes = tf.transpose(state, index_permutation)

    return reordered_modes


def insert_state(state, system, state_is_pure, mode=None, batched=False):
    """
    Append a new mode (at slot 'mode') to system and initialize it in 'state'.
    If 'mode' is not specified or is greater than the largest current mode number, the new mode is added to the end.
    If an integer within [0,...,N-1] is given for 'mode' (where N is the number of modes) in 'system,'
    then the new state is put in the corresponding mode, and all following modes are shifted to the right by one.
    """
    # pylint: disable=too-many-branches
    num_indices = len(system.shape)
    if batched:
        batch_offset = 1
    else:
        batch_offset = 0
    if state_is_pure:
        num_modes = num_indices - batch_offset
    else:
        num_modes = (num_indices - batch_offset) // 2

    if mode is None or mode >= num_modes:
        mode = num_modes

    if len(system.shape) == 0:  # pylint: disable=len-as-condition
        # no modes in system
        # pylint: disable=no-else-return
        if len(state.shape) - batch_offset == 1:
            if state_is_pure:
                return state
            else:
                return mixed(state)
        elif len(state.shape) - batch_offset == 2:
            return state
        else:
            raise ValueError(
                "'state' must have dim={} or dim={}".format(1 - batch_offset, 2 - batch_offset)
            )
    else:
        # modes in system
        if len(state.shape) == batch_offset + 1 and state_is_pure:
            # everything is pure
            # basic form:
            # 'ab...ln...yz,m->ab...lmn...yz' ('m' indices belong to bra of mode being inserted)
            mode_size = 1
        else:
            # everything is mixed
            # basic form:
            # 'abcd...klop...wxyz,mn->abcd...klmnop...wxyz' ('mn' indices belong to bra/ket of mode being inserted)
            mode_size = 2

        batch_index = indices[:batch_offset]
        left_part = indices[batch_offset : batch_offset + mode * mode_size]
        middle_part = indices[
            batch_offset + mode * mode_size : batch_offset + (mode + 1) * mode_size
        ]
        right_part = indices[
            batch_offset + (mode + 1) * mode_size : batch_offset + (num_modes + 1) * mode_size
        ]
        eqn_lhs = batch_index + left_part + right_part + "," + batch_index + middle_part
        eqn_rhs = batch_index + left_part + middle_part + right_part
        eqn = eqn_lhs + "->" + eqn_rhs
        revised_modes = tf.einsum(eqn, system, state)

    return revised_modes


def partial_trace(system, mode, state_is_pure, batched=False):
    """
    Trace out subsystem 'mode' from 'system'.
    This operation always returns a mixed state, since we do not know in advance if a mode is entangled with others.
    """
    num_indices = len(system.shape)
    if batched:
        batch_offset = 1
    else:
        batch_offset = 0
    if state_is_pure:
        num_modes = num_indices - batch_offset
    else:
        num_modes = (num_indices - batch_offset) // 2

    if state_is_pure:
        system = mixed(system, batched)

    # tensorflow trace implementation
    # requires subsystem to be traced out to be at end
    # i.e., ab...klmnop...yz goes to ab...klop...yzmn
    indices_list = [m for m in range(batch_offset + 2 * num_modes)]
    dim_list = (
        indices_list[: batch_offset + 2 * mode]
        + indices_list[batch_offset + 2 * (mode + 1) :]
        + indices_list[batch_offset + 2 * mode : batch_offset + 2 * (mode + 1)]
    )
    permuted_sys = tf.transpose(system, dim_list)
    reduced_state = tf.trace(permuted_sys)
    return reduced_state


def reduced_density_matrix(system, mode, state_is_pure, batched=False):
    """
    Trace out all subsystems except 'mode' from 'system'.
    This operation always returns a mixed state, since we do not know in advance if a mode is entangled with others.
    """
    if state_is_pure:
        reduced_state = mixed(system, batched)
    else:
        reduced_state = system
    num_indices = len(reduced_state.shape)
    if batched:
        batch_offset = 1
    else:
        batch_offset = 0
    num_modes = (num_indices - batch_offset) // 2  # always mixed
    for m in range(num_modes):
        if m != mode:
            reduced_state = partial_trace(reduced_state, m, False, batched)
    return reduced_state


def conditional_state(system, projector, mode, state_is_pure, batched=False):
    """Compute the (unnormalized) conditional state of 'system' after applying ket 'projector' to 'mode'."""
    # basic_form (pure states): abc...ijk...xyz,j-> abc...ik...xyz
    # basic_form (mixed states): abcd...ijklmn...wxyz,k,l-> abcd...ijmn...wxyz
    mode = int(mode)
    if not batched:
        # add in fake batch dimension (TF 1.5 gives unexplainable errors if we do it without this)
        system = tf.expand_dims(system, 0)
        projector = tf.expand_dims(projector, 0)
    num_indices = len(system.shape)
    batch_offset = 1
    if state_is_pure:
        mode_size = 1
    else:
        mode_size = 2
    num_modes = (num_indices - batch_offset) // mode_size
    max_num = (max_num_indices - batch_offset) // num_modes
    if num_modes > max_num:
        raise ValueError(
            "Conditional state projection currently only supported for {} modes.".format(max_num)
        )

    batch_index = indices[:batch_offset]
    mode_indices = indices[batch_offset : batch_offset + num_modes * mode_size]
    projector_indices = mode_indices[:mode_size]
    free_mode_indices = mode_indices[mode_size : num_modes * mode_size]
    state_lhs = (
        batch_index
        + free_mode_indices[: mode * mode_size]
        + projector_indices
        + free_mode_indices[mode * mode_size :]
    )
    projector_lhs = batch_index + projector_indices[0]
    if mode_size == 2:
        projector_lhs += "," + batch_index + projector_indices[1]
    eqn_lhs = ",".join([state_lhs, projector_lhs])
    eqn_rhs = batch_index + free_mode_indices
    eqn = eqn_lhs + "->" + eqn_rhs
    einsum_args = [system, tf.conj(projector)]
    if not state_is_pure:
        einsum_args.append(projector)
    cond_state = tf.einsum(eqn, *einsum_args)
    if not batched:
        cond_state = tf.squeeze(cond_state, 0)  # drop fake batch dimension
    return cond_state


###################################################################

# Helpful auxiliary functions for ops


def ladder_ops(D):
    """returns the matrix representation of the annihilation and creation operators"""
    ind = [(i - 1, i) for i in range(1, D)]
    updates = [np.sqrt(i) for i in range(1, D)]
    shape = [D, D]
    a = tf.scatter_nd(ind, updates, shape)
    ad = tf.transpose(tf.conj(a), [1, 0])
    return a, ad


def H(n, x):
    """Explicit expression for Hermite polynomials."""
    prefactor = factorial(n)
    terms = tf.reduce_sum(
        tf.stack(
            [
                _numer_safe_power(-1, m)
                / (factorial(m) * factorial(n - 2 * m))
                * _numer_safe_power(2 * x, n - 2 * m)
                for m in range(int(np.floor(n / 2)) + 1)
            ],
            axis=0,
        ),
        axis=0,
    )
    return prefactor * terms


def H_n_plus_1(H_n, H_n_m1, n, x):
    """Recurrent definition of Hermite polynomials."""
    H_n_p1 = 2 * x * H_n - 2 * n * H_n_m1
    return H_n_p1


###################################################################

# Helper functions


def _check_for_eval(kwargs):
    """
    Helper function to check keyword arguments for user-supplied information about how numerical evaluation should proceed,
    what session to use, and what the feed dictionary should be.
    """
    if "feed_dict" in kwargs:
        feed_dict = kwargs["feed_dict"]
    else:
        feed_dict = {}
    if "session" in kwargs:
        # evaluate tensors in supplied session
        session = kwargs["session"]
        evaluate_results = True
        close_session = False
    elif "eval" in kwargs and kwargs["eval"] is False:
        # don't evaluate tensors
        session = None
        evaluate_results = False
        close_session = False
    else:
        # evaluate tensors in temporary session
        session = tf.Session()
        evaluate_results = True
        close_session = True
    return evaluate_results, session, feed_dict, close_session



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
A Strawberry Fields module that provides an object-oriented interface for building
quantum circuit representations of continuous-variable circuits using the
:math:`\LaTeX` `Qcircuit package <https://ctan.org/pkg/qcircuit>`_.
"""
import datetime
import os

# string constants used by :class:`~strawberryfields.circuitdrawer.Circuit`
# to transform Strawberry Fields operators to latex code.

DOCUMENT_CLASS = r"\documentclass{article}"
EMPTY_PAGESTYLE = r"\pagestyle{empty}"
QCIRCUIT_PACKAGE = r"\usepackage{qcircuit}"
BEGIN_DOCUMENT = r"\begin{document}"
DOCUMENT_END = r"\end{document}"
CIRCUIT_START = r"\Qcircuit"
COLUMN_SPACING = "@C={0}"  # spacing i.e. "1em"
ROW_SPACING = "@R={0}"  # spacing
UNIFORM_ROW_SPACING = "@!R"
UNIFORM_COLUMN_SPACING = "@!C"
UNIFORM_ELEMENT_SPACING = "@!"
QUANTUM_WIRE = r"\qw"
MULTI_QUANTUM_WIRE = r"\qw[{0}]"  # length vector i.e. "-1"
VERTICAL_QUANTUM_WIRE = r"\qwx[{0}]"  # length vector
WIRE_END = r"\qwa[{0}]"  # length vector
CLASSICAL_WIRE = r"\cw[{0}]"  # length vector
CLASSICAL_WIRE_END = r"\cwa[{0}]"  # length vector
VERTICAL_CLASSICAL_WIRE = r"\cwx[{0}]"  # length vector
LABELLED_GATE = r"\gate{{{0}}}"  # label i.e. "X"
TARGET = r"\targ"
SWAP = r"\qswap"
MULTIGATE = r"\multigate{{{0}}}{{{1}}}"  # depth i.e. "2", label
NON_ADJACENT_MULTIGATE = r"\sgate{{{0}}}{{{1}}}"  # gate i.e. "X", second wire vector i.e. "1"
GHOST = r"\ghost{{{0}}}"  # gate
CLASSICAL_GHOST = r"\cghost{{{0}}}"  # gate
NO_GHOST = r"\nghost{{{0}}}"  # gate
CONTROL = r"\ctrl{{{0}}}"  # length vector
CONTROL_ON_ZERO = r"\ctrlo{{{0}}}"  # length vector
CLASSICAL_CONTROL = r"\cctrl{{{0}}}"  # length vector
CLASSICAL_CONTROL_ON_ZERO = r"\cctrlo{{{0}}}"  # length vector
ISOLATED_CONTROL = r"\control"
ISOLATED_CONTROL_ON_ZERO = r"\controlo"
METER = r"\meter"
BASIS_METER = r"\meterB{{{0}}}"  # basis i.e. \ket{\xi_\pm}
SPLIT_BASIS_METER = r"\smeterB{{{0}}}{{{1}}}"  # basis, second wire vector
MEASURE = r"\measuretab{{{0}}}"  # label
MULTIMEASURE = r"\multimeasure{{{0}}}{{{1}}}"  # depth, label
LEFT_WIRE_LABEL = r"\lstick{{{0}}}"  # label
RIGHT_WIRE_LABEL = r"\rstick{{{0}}}"  # label
BRA = r"\bra{{{0}}}"  # state i.e. "\psi"
KET = r"\ket{{{0}}}"  # state

HADAMARD_COMP = LABELLED_GATE.format("H")
PAULI_X_COMP = LABELLED_GATE.format("X")
PAULI_Y_COMP = LABELLED_GATE.format("Y")
PAULI_Z_COMP = LABELLED_GATE.format("Z")
D_COMP = LABELLED_GATE.format("D")
S_COMP = LABELLED_GATE.format("S")
R_COMP = LABELLED_GATE.format("R")
P_COMP = LABELLED_GATE.format("P")
V_COMP = LABELLED_GATE.format("V")
K_COMP = LABELLED_GATE.format("K")
FOURIER_COMP = LABELLED_GATE.format("F")

BS_MULTI_COMP = "BS"
S_MULTI_COMP = "S"

WIRE_OPERATION = "& {0}"
WIRE_TERMINATOR = r"\\" + "\n"
CIRCUIT_BODY_TERMINATOR = "}\n"
CIRCUIT_BODY_START = " {" + "\n"
INIT_DOCUMENT = (
    DOCUMENT_CLASS
    + "\n"
    + EMPTY_PAGESTYLE
    + "\n"
    + QCIRCUIT_PACKAGE
    + "\n"
    + BEGIN_DOCUMENT
    + "\n"
    + CIRCUIT_START
)

PIPE = "|"
LINE_RETURN = "\n"


class NotDrawableException(Exception):
    """Exception raised when a circuit is not drawable.

    This class corresponds to the exception raised by :meth:`~.parse_op`
    when a circuit is deemed impossible to effectively render using qcircuit.
    """

    pass


class ModeMismatchException(Exception):
    """Exception raised when parsing a Gate object.

    This class corresponds to the exception raised by :meth:`~.parse_op`
    when an operator is interpreted as an n-mode gate but is applied to a number of modes != n.
    """

    pass


class UnsupportedGateException(Exception):
    """Exception raised when attempting to add an unsupported operator.

    This class corresponds to the exception raised by :meth:`~.parse_op` when it is attempted to add an
    unsupported operator to the circuit.
    """

    pass


class Circuit:
    """Represents a quantum circuit that can be compiled to tex format.

    Args:
        wires (int): the number of quantum wires or subsystems to use in the
            circuit diagram.
    """

    _circuit_matrix = []

    def __init__(self, wires):
        self._document = ""
        self._circuit_matrix = [[QUANTUM_WIRE.format(1)] for wire in range(wires)]
        self._column_spacing = None
        self._row_spacing = None

        self.single_mode_gates = {
            "Xgate": self._x,
            "Zgate": self._z,
            "Dgate": self._d,
            "Sgate": self._s,
            "Rgate": self._r,
            "Pgate": self._p,
            "Vgate": self._v,
            "Kgate": self._k,
            "Fourier": self._fourier,
        }

        self.two_mode_gates = {
            "CXgate": self._cx,
            "CZgate": self._cz,
            "CKgate": self._ck,
            "BSgate": self._bs,
            "S2gate": self._s2,
        }

    def _gate_from_operator(self, op):
        """Infers the number of modes and callable Circuit class method that correspond
        with a Strawberry Fields operator object.

        Args:
            op (strawberryfields.ops.Gate): the Strawberry Fields operator object.

        Returns:
            method (function): callable method that adds the given operator to the latex circuit.
            mode (int): the number of modes affected by the operator gate.
        """
        operator = str(op).split(PIPE)[0]
        method = None
        mode = None

        for two_mode_gate in self.two_mode_gates:
            if two_mode_gate in operator:
                method = self.two_mode_gates[two_mode_gate]
                mode = 2

        if method is None:
            for single_mode_gate in self.single_mode_gates:
                if single_mode_gate in operator:
                    method = self.single_mode_gates[single_mode_gate]
                    mode = 1

        return method, mode

    def parse_op(self, op):
        """Transforms a Strawberry Fields operator object to a latex qcircuit gate.

        Args:
            op (strawberryfields.ops.Gate): the Strawberry Fields operator object.

        Raises:
            UnsupportedGateException: if the operator is not supported by the circuit drawer module.
            ModeMismatchException: if the operator is interpreted as an n-mode gate but is
                applied to a number of modes != n.

        """
        if not op.__class__.__name__ == "Command":
            return

        method, mode = self._gate_from_operator(op)
        wires = list(map(lambda register: register.ind, op.reg))

        if method is None:
            raise UnsupportedGateException(
                "Unsupported operation {0} not printable by circuit builder!".format(str(op))
            )
        if mode == len(wires):
            method(*wires)
        else:
            raise ModeMismatchException(
                "{0} mode gate applied to {1} wires!".format(mode, len(wires))
            )

    def _x(self, wire):
        """Adds a position displacement operator to the circuit.

        Args:
            wire (int): the subsystem wire to apply the operator to.
        """
        self._single_mode_gate(wire, PAULI_X_COMP)

    def _z(self, wire):
        """Adds a momentum displacement operator to the circuit.

        Args:
            wire (int): the subsystem wire to apply the operator to.
        """
        self._single_mode_gate(wire, PAULI_Z_COMP)

    def _s(self, wire):
        """Adds a squeezing operator to the circuit.

        Args:
            wire (int): the subsystem wire to apply the operator to.
        """
        self._single_mode_gate(wire, S_COMP)

    def _d(self, wire):
        """Adds a displacement operator to the circuit.

        Args:
            wire (int): the subsystem wire to apply the operator to.
        """
        self._single_mode_gate(wire, D_COMP)

    def _r(self, wire):
        """Adds a rotation operator to the circuit.

        Args:
            wire (int): the subsystem wire to apply the operator to.
        """
        self._single_mode_gate(wire, R_COMP)

    def _p(self, wire):
        """Adds a quadratic phase shift operator to the circuit.

        Args:
            wire (int): the subsystem wire to apply the operator to.
        """
        self._single_mode_gate(wire, P_COMP)

    def _v(self, wire):
        """Adds a cubic phase shift operator to the circuit.

        Args:
            wire (int): the subsystem wire to apply the operator to.
        """
        self._single_mode_gate(wire, V_COMP)

    def _k(self, wire):
        """Adds a Kerr operator to the circuit.

        Args:
            wire (int): the subsystem wire to apply the operator to.
        """
        self._single_mode_gate(wire, K_COMP)

    def _fourier(self, wire):
        """Adds a Fourier transform operator to the circuit.

        Args:
            wire (int): the subsystem wire to apply the operator to.
        """
        self._single_mode_gate(wire, FOURIER_COMP)

    def _cx(self, source_wire, target_wire):
        """Adds a controlled position displacement operator to the circuit.

        Args:
            source_wire (int): the controlling subsystem wire.
            target_wire (int): the controlled subsystem wire.
        """
        self._controlled_mode_gate(source_wire, target_wire, TARGET)

    def _cz(self, source_wire, target_wire):
        """Adds a controlled phase operator to the circuit.

        Args:
            source_wire (int): the controlling subsystem wire.
            target_wire (int): the controlled subsystem wire.
        """
        self._controlled_mode_gate(source_wire, target_wire, PAULI_Z_COMP)

    def _ck(self, source_wire, target_wire):
        """Adds a controlled Kerr operator to the circuit.

        Args:
            source_wire (int): the controlling subsystem wire.
            target_wire (int): the controlled subsystem wire.
        """
        self._controlled_mode_gate(source_wire, target_wire, K_COMP)

    def _bs(self, first_wire, second_wire):
        """Adds a beams plitter operator to the circuit.

        Args:
            first_wire (int): the first subsystem wire to apply the operator to.
            second_wire (int): the second subsystem wire to apply the operator to.
        """
        self._multi_mode_gate(BS_MULTI_COMP, [first_wire, second_wire])

    def _s2(self, first_wire, second_wire):
        """Adds an two mode squeezing operator to the circuit.

        Args:
            first_wire (int): the first subsystem wire to apply the operator to.
            second_wire (int): the second subsystem wire to apply the operator to.
        """
        self._multi_mode_gate(S_MULTI_COMP, [first_wire, second_wire])

    # operation types

    def _single_mode_gate(self, wire, circuit_op):
        """Adds a single-mode operator gate to the circuit.

        Args:
            circuit_op (str): the latex code for the operator.
            wires (list[int]): a list of the indeces of subsystem wires to apply the multi-mode gate to.
        """
        matrix = self._circuit_matrix
        wire_ops = matrix[wire]

        if Circuit._is_empty(wire_ops[-1]):
            wire_ops[-1] = circuit_op
        else:
            wire_ops.append(circuit_op)
            for prev_wire in matrix[:wire]:
                prev_wire.append(QUANTUM_WIRE.format(1))
            for post_wire in matrix[wire + 1 :]:
                post_wire.append(QUANTUM_WIRE.format(1))

    def _multi_mode_gate(self, circuit_op, wires):
        """Adds a multi-mode operator to the circuit.

        Args:
            circuit_op (str): the latex code for the operator.
            wires (list[int]): a list of the indeces of subsystem wires to apply the gate to.
        Raises:
            ModeMismatchException: if the operator is applied to non-adjacent wires.
        """
        matrix = self._circuit_matrix

        if not self._on_empty_column():
            self._add_column()

        wires.sort()

        first_wire = wires.pop(0)
        wire_ops = matrix[first_wire]
        wire_ops[-1] = MULTIGATE.format(1, circuit_op)
        matrix[first_wire] = wire_ops
        previous_wire = first_wire

        for wire in wires:
            if not previous_wire == wire - 1:
                raise NotDrawableException(
                    "{0} multi-mode gate applied to non-adjacent wires!".format(circuit_op)
                )
            wire_ops = matrix[wire]
            wire_ops[-1] = GHOST.format(circuit_op)
            matrix[wire] = wire_ops
            previous_wire = wire

        self._circuit_matrix = matrix

    def _controlled_mode_gate(self, source_wire, target_wire, circuit_op):
        """Adds a controlled operator gate to the circuit.

        Args:
            source wire (int): the index of the controlling subsystem.
            target_wire (int): the index of the controlled subsystem.
            circuit_op (str): the latex code for the operator.
        """
        matrix = self._circuit_matrix
        source_ops = matrix[source_wire]
        target_ops = matrix[target_wire]
        distance = target_wire - source_wire

        if Circuit._is_empty(source_ops[-1]) and Circuit._is_empty(target_ops[-1]):
            source_ops[-1] = CONTROL.format(distance)
            target_ops[-1] = circuit_op
        else:
            for index, wire_ops in enumerate(matrix):
                if index == source_wire:
                    wire_ops.append(CONTROL.format(distance))
                elif index == target_wire:
                    wire_ops.append(circuit_op)
                else:
                    wire_ops.append(QUANTUM_WIRE.format(1))

    # helpers

    def _on_empty_column(self):
        """Checks if the right-most wires for each subsystem in the circuit are all empty

        Returns:
            bool: whether the right-most wires for each subsystem in the circuit are all empty
        """
        matrix = self._circuit_matrix

        empty_column = True
        for wire in enumerate(matrix):
            wire_ops = wire[1]
            if not Circuit._is_empty(wire_ops[-1]):
                empty_column = False
                break

        return empty_column

    def _add_column(self):
        """Adds a unit of quantum wire to each subsystem in the circuit."""
        for wire in self._circuit_matrix:
            wire.append(QUANTUM_WIRE.format(1))

    @staticmethod
    def _is_empty(op):
        """Checks for a NOP, a quantum wire location without an operator.

        Args:
            op (str): latex code for either an operator, or empty quantum wire.

        Returns:
            bool: whether the argument is an empty quantum wire.
        """
        return op == QUANTUM_WIRE.format(1)

    # cosmetic

    def _set_column_spacing(self, spacing):
        """Sets visual spacing between operators in quantum circuit.

        Args:
            spacing (int): spacing between operators.
        """
        self._column_spacing = spacing

    def _set_row_spacing(self, spacing):
        """Sets visual spacing of wires in quantum circuit.

        Args:
            spacing (int): spacing between wires.
        """
        self._row_spacing = spacing

    @staticmethod
    def _pad_with_spaces(string):
        """Pads string with spaces.

        Args:
            string (str): string to pad.

        Returns:
            str: string with space added to either side.
        """
        return " " + string + " "

    # latex translation

    def dump_to_document(self):
        """Writes current circuit to document.

        Returns:
            str: latex document string.
        """
        self._init_document()
        self._apply_spacing()
        self._begin_circuit()

        self._add_column()

        for wire_ops in enumerate(self._circuit_matrix):
            for wire_op in wire_ops[1]:
                self._write_operation_to_document(wire_op)
            self._end_wire()

        self._end_circuit()
        self._end_document()

        return self._document

    def compile_document(self, tex_dir="./circuit_tex"):
        """Compiles latex documents.

        Args:
            tex_dir (str): relative directory for latex document output.

        Returns:
            str: the file path of the resulting latex document.
        """
        tex_dir = os.path.abspath(tex_dir)

        if not os.path.isdir(tex_dir):
            os.mkdir(tex_dir)

        file_name = "output_{0}".format(datetime.datetime.now().strftime("%Y_%B_%d_%I:%M%p"))
        file_path = "{0}/{1}.tex".format(tex_dir, file_name)

        with open(file_path, "w+") as output_file:
            output_file.write(self._document)

        return file_path

    def _init_document(self):
        """Adds the required latex headers to the document."""
        self._document = INIT_DOCUMENT

    def _end_document(self):
        """Appends latex EOD code to the document."""
        self._document += DOCUMENT_END

    def _begin_circuit(self):
        """Prepares document for latex circuit content."""
        self._document += CIRCUIT_BODY_START

    def _end_circuit(self):
        """Ends the latex circuit content."""
        self._document += CIRCUIT_BODY_TERMINATOR

    def _end_wire(self):
        """Ends a wire within the latex circuit."""
        self._document += WIRE_TERMINATOR

    def _apply_spacing(self):
        """Applies wire and operator visual spacing."""
        if self._column_spacing is not None:
            self._document += Circuit._pad_with_spaces(COLUMN_SPACING.format(self._column_spacing))
        if self._row_spacing is not None:
            self._document += Circuit._pad_with_spaces(ROW_SPACING.format(self._row_spacing))

    def _write_operation_to_document(self, operation):
        """Appends operation latex code to circuit in latex document.

        Args:
            operation (str): the latex code for the quantum operation to be applied.
        """
        self._document += Circuit._pad_with_spaces(WIRE_OPERATION.format(operation))

    def __str__(self):
        """String representation of the Circuit class."""
        return self._document



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This subpackage implements the :class:`CircuitSpecs` class, an abstract base class
used to define classes or families of quantum circuits, e.g., circuits that can be executed on particular
hardware or simulator backends.

The information in the :class:`CircuitSpecs` instances is used by :meth:`.Program.compile` to validate and
compile quantum programs. By querying the :class:`CircuitSpecs` class representing the requested compilation
target, :meth:`.Program.compile` can

1. **Validate** that the Program has the correct number of modes, and consists
   of valid quantum operations in the correct topology for the targeted circuit class.

2. **Compile** the Program into an :term:`equivalent circuit` that has the topology required by the
   targeted circuit class, decomposing circuit operations as required.

Note that the compilation process is not perfect and can provide false negatives: it can admit
failure by raising a :class:`.CircuitError` even if the Program theoretically is equivalent to a
circuit that belongs in the target circuit class.

The circuit class database :attr:`circuit_db` is a dictionary mapping the circuit family
short name to the corresponding CircuitSpecs instance.
In particular, for each backend supported by Strawberry Fields the database contains a
corresponding CircuitSpecs instance with the same short name, used to validate Programs to be
executed on that backend.
"""
from .circuit_specs import CircuitSpecs
from .X8 import X8Specs, X8_01
from .X12 import X12Specs, X12_01, X12_02
from .fock import FockSpecs
from .gaussian import GaussianSpecs
from .gbs import GBSSpecs
from .tensorflow import TFSpecs
from .gaussian_unitary import GaussianUnitary

specs = (
    X8Specs,
    X8_01,
    X12Specs,
    X12_01,
    X12_02,
    FockSpecs,
    GaussianSpecs,
    GBSSpecs,
    TFSpecs,
    GaussianUnitary,
)

circuit_db = {c.short_name: c for c in specs}
"""dict[str, ~strawberryfields.circuitspecs.CircuitSpecs]: Map from circuit
family short name to the corresponding class."""

__all__ = ["circuit_db", "CircuitSpecs"] + [i.__name__ for i in specs]



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common shared operations that can be used by backends"""

import os
import functools
import re
import itertools
from bisect import bisect
import pkg_resources

import numpy as np
import scipy as sp
from scipy.special import binom
from scipy.special import gammaln as lg
from scipy.linalg import qr

DATA_PATH = pkg_resources.resource_filename("strawberryfields", "backends/data")
def_type = np.complex128


# ================================+
#   Fock space shared operations |
# ================================+


@functools.lru_cache()
def find_dim_files(regex, D, directory=None, name=""):
    r"""Return files matching a certain regex for specified dimension D.

    If no such file satisfying this condition exists, then an Exception
    is returned.

    Args:
        regex (str): regex matching the allowed file names. Should contain
        "(\d+)", this represents the Fock dimension in the file name.
        D (int): the dimension D
        directory (str): location to load the precomputed beamsplitter
            factors from. By default, this will be the Strawberry Fields data directory.
    """
    if directory is None:
        directory = DATA_PATH
    else:
        check_dir = os.path.isdir(directory)
        if not check_dir:  # pragma: no cover
            raise NotADirectoryError("Directory {} does not exist!".format(directory))

    files = [f for f in os.listdir(directory) if re.match(regex, f)]
    avail_dims = sorted([int(re.findall(r"\d+", f)[0]) for f in files])

    idx = bisect(avail_dims, D - 1)
    if idx + 1 > len(avail_dims):
        raise FileNotFoundError(
            "File containing {} factors does not exist "
            "for dimension {} in directory {}".format(name, D, directory)
        )

    return avail_dims[idx], directory


@functools.lru_cache()
def generate_bs_factors(D):
    r"""Generate beamsplitter factors in the Fock basis.

    This function generates the beamsplitter prefactors,

        .. math::
            prefac_{N,n,M,m,k} = (-1)^{N-k}\sqrt{\binom{n,k}\binom{m,N-k}\binom{N,k}\binom{M,n-k}}

    for a specific cutoff dimension :math:`D`.

    Note that the last dimension will only contain non-zero values
    for indices ``0`` to ``n``.

    Args:
        D (int): generate prefactors for :math:`D` dimensions.
    """
    prefac = np.zeros([D] * 5, dtype=def_type)

    for (N, M, n) in itertools.product(*([range(D)] * 3)):
        m = N + M - n
        k = np.arange(n + 1)
        if 0 <= m < D:
            # pylint: disable=bad-whitespace
            prefac[N, n, M, m, : n + 1] = (-1.0) ** (N - k) * np.sqrt(
                binom(n, k) * binom(m, N - k) * binom(N, k) * binom(M, n - k)
            )

    return prefac


@functools.lru_cache()
def load_bs_factors(D, directory=None):
    r"""Load precomputed beamsplitter factors in the Fock basis.

    This function searches the data directory for a BS prefactor file
    containing for cutoff dimension higher or equal to that requested
    (``D``). It then reshapes the rank-2 sparse array to a
    :math:`D\times D\times D\times D\times D` dense array.

    If no such file satisfying this condition exists, then an Exception
    is returned.

    Args:
        D (int): load prefactors containing at least ``D`` dimensions.
        directory (str): location to load the precomputed beamsplitter
            factors from. By default, this will be the Strawberry Fields data directory.
    """

    regex = r"fock_beamsplitter_factors_(\d+)\.npz"
    load_dim, location = find_dim_files(regex, D, directory=directory, name="beamsplitter")
    filename = "fock_beamsplitter_factors_{}.npz".format(load_dim)
    prefac = sp.sparse.load_npz(os.path.join(location, filename))
    return np.reshape(prefac.toarray(), [load_dim] * 5)


def save_bs_factors(prefac, directory=None):
    r"""Saves precomputed beamsplitter factors in the Fock basis to a file.

    This function reshapes the rank-5 array with dimension
    :math:`D\times D\times D\times D\times D` to a rank-2 array of dimension
    :math:`D^4\times D`, before converting it to a sparse array, and saving
    it to a file in the specified directory.

    Args:
        prefac (numpy.array): the Numpy array containing the precomputed beamsplitter
            prefactors in the Fock basis. Must be of size [D,D,D,D,D] for some integer D
        directory (str): location to save the precomputed beamsplitter factors. By default,
            this will be the Strawberry Fields data directory.
    """
    if directory is None:
        directory = DATA_PATH
    else:
        check_dir = os.path.isdir(directory)
        if not check_dir:  # pragma: no cover
            raise NotADirectoryError("Directory {} does not exist!".format(directory))

    D = prefac.shape[0]
    filename = "fock_beamsplitter_factors_{}.npz".format(D)

    prefac_rank2 = np.reshape(prefac, ((D) ** 4, D), order="C")
    prefac_sparse = sp.sparse.csc_matrix(prefac_rank2)

    sp.sparse.save_npz(os.path.join(directory, filename), prefac_sparse)


@functools.lru_cache()
def squeeze_parity(D):
    r"""Creates the parity prefactor needed for squeezing in the Fock basis.

    .. math::
        \text{\sigma}_{N,k} = \begin{cases}
            (N-k)/2, & \text{mod}(N-k,2) \neq 0\\
            0, &\text{otherwise}
        \end{cases}

    Args:
        D (numpy.array): generate the prefactors for a Fock truncation of :math:`D`.
    """
    k = np.int(np.ceil(D / 4) * 4)
    v = np.full(k, 1)
    v[1::2] = 0
    v[2::4] = -1
    v = np.vstack([np.roll(v, i) for i in range(k)])
    return v[:D, :D]


@functools.lru_cache()
def generate_squeeze_factors(D):
    r"""Generate squeezing factors in the Fock basis.

    This function generates the squeezing prefactors,

        .. math::
            prefac_{N,n,k} = \frac{\sigma_{N,k}\sqrt{n!N!}}
            {k!\left(\frac{n-k}{2}\right)!\left(\frac{N-k}{2}\right)!}

    where :math:`\sigma_{N,k}` is the parity, given by :func:`~.squeeze_parity`.

    Args:
        D (int): generate prefactors for :math:`D` dimensions.
    """
    dim_array = np.arange(D)
    N = dim_array.reshape((-1, 1, 1))
    n = dim_array.reshape((1, -1, 1))
    k = dim_array.reshape((1, 1, -1))

    # we only perform the sum when n+N is divisible by 2
    # in which case we sum 0 <= k <= min(N,n)
    mask = np.logical_and((n + N) % 2 == 0, k <= np.minimum(N, n))

    # need to use np.power to avoid taking the root of a negative
    # in the numerator (these are discarded by the mask anyway)
    signs = squeeze_parity(D).reshape([D, 1, D])
    logfac = np.where(
        mask,
        0.5 * (lg(n + 1) + lg(N + 1)) - lg(k + 1) - lg((n - k) / 2 + 1) - lg((N - k) / 2 + 1),
        0,
    )

    if D <= 600:
        prefactor = np.exp(logfac, dtype=np.float64) * signs * mask
    else:
        prefactor = np.exp(logfac, dtype=np.float128) * signs * mask

    return prefactor


def save_squeeze_factors(prefac, directory=None):
    r"""Saves precomputed squeeze factors in the Fock basis to a file.

    This function reshapes the rank-3 array with dimension
    :math:`D\times D\times D` to a rank-2 array of dimension
    :math:`D^2\times D`, before converting it to a sparse array, and saving
    it to a file in the specified directory.

    Args:
        prefac (numpy.array): the Numpy array containing the precomputed squeeze
            prefactors in the Fock basis. Must be of size [D,D,D] for some integer D
        directory (str): location to save the precomputed beamsplitter factors. By default,
            this will be the Strawberry Fields data directory.
    """
    if directory is None:
        directory = DATA_PATH
    else:
        check_dir = os.path.isdir(directory)
        if not check_dir:
            raise NotADirectoryError("Directory {} does not exist!".format(directory))

    D = prefac.shape[0]
    filename = "fock_squeeze_factors_{}.npz".format(D)

    prefac_rank2 = np.reshape(prefac, ((D) ** 2, D), order="C")
    prefac_sparse = sp.sparse.csc_matrix(prefac_rank2)

    sp.sparse.save_npz(os.path.join(directory, filename), prefac_sparse)


@functools.lru_cache()
def load_squeeze_factors(D, directory=None):
    r"""Load precomputed squeeze factors in the Fock basis.

    This function searches the data directory for a squeeze prefactor file
    containing for cutoff dimension higher or equal to that requested
    (``D``). It then reshapes the rank-2 sparse array to a
    :math:`D\times D\times D` dense array.

    If no such file satisfying this condition exists, then an Exception
    is returned.

    Args:
        D (int): load prefactors containing at least ``D`` dimensions.
        directory (str): location to load the precomputed squeeze
            factors from. By default, this will be the Strawberry Fields data directory.
    """
    regex = r"fock_squeeze_factors_(\d+)\.npz"
    load_dim, location = find_dim_files(regex, D, directory=directory, name="squeeze")

    filename = "fock_squeeze_factors_{}.npz".format(load_dim)
    prefac = sp.sparse.load_npz(os.path.join(location, filename))

    return np.reshape(prefac.toarray(), [load_dim] * 3)


# ================================+
# Phase space shared operations  |
# ================================+


@functools.lru_cache()
def rotation_matrix(phi):
    r"""Rotation matrix.

    Args:
        phi (float): rotation angle
    Returns:
        array: :math:`2\times 2` rotation matrix
    """
    return np.array([[np.cos(phi), -np.sin(phi)], [np.sin(phi), np.cos(phi)]])


@functools.lru_cache()
def sympmat(n):
    r""" Returns the symplectic matrix of order n

    Args:
        n (int): order
        hbar (float): the value of hbar used in the definition
            of the quadrature operators
    Returns:
        array: symplectic matrix
    """
    idm = np.identity(n)
    omega = np.concatenate(
        (np.concatenate((0 * idm, idm), axis=1), np.concatenate((-idm, 0 * idm), axis=1)), axis=0
    )
    return omega


@functools.lru_cache()
def changebasis(n):
    r"""Change of basis matrix between the two Gaussian representation orderings.

    This is the matrix necessary to transform covariances matrices written
    in the (x_1,...,x_n,p_1,...,p_n) to the (x_1,p_1,...,x_n,p_n) ordering

    Args:
        n (int): number of modes
    Returns:
        array: :math:`2n\times 2n` matrix
    """
    m = np.zeros((2 * n, 2 * n))
    for i in range(n):
        m[2 * i, i] = 1
        m[2 * i + 1, i + n] = 1
    return m


def haar_measure(n):
    """A Random matrix distributed with the Haar measure.

    For more details, see :cite:`mezzadri2006`.

    Args:
        n (int): matrix size
    Returns:
        array: an nxn random matrix
    """
    z = (sp.randn(n, n) + 1j * sp.randn(n, n)) / np.sqrt(2.0)
    q, r = qr(z)
    d = sp.diagonal(r)
    ph = d / np.abs(d)
    q = np.multiply(q, ph, q)
    return q



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Tools for visualizing graphs, subgraphs, point processes, and vibronic spectra.

Visualization requires installation of the Plotly library, which is not a dependency of
Strawberry Fields. Plotly can be installed using ``pip install plotly`` or by visiting their
`installation instructions <https://plot.ly/python/getting-started/#installation>`__.
"""
# pylint: disable=import-outside-toplevel
from typing import Optional, Tuple

import networkx as nx
import numpy as np


def _node_coords(g: nx.Graph, l: dict) -> Tuple:
    """Converts coordinates for the graph nodes for plotting purposes.

    Args:
        g (nx.Graph): input graph
        l (dict[int, float]): Dictionary of nodes and their respective coordinates. Can be
            generated using a NetworkX `layout <https://networkx.github.io/documentation/latest/
            reference/drawing.html#module-networkx.drawing.layout>`__

    Returns:
         dict[str, list]: lists of x and y coordinates accessed as keys of a dictionary
    """
    n_x = []
    n_y = []

    for n in g.nodes():
        n_x.append(l[n][0])
        n_y.append(l[n][1])

    return {"x": n_x, "y": n_y}


def _edge_coords(g: nx.Graph, l: dict) -> dict:
    """Converts coordinates for the graph edges for plotting purposes.

        Args:
            g (nx.Graph): input graph
            l (dict[int, float]): Dictionary of nodes and their respective coordinates. Can be
                generated using a NetworkX `layout <https://networkx.github.io/documentation/latest/
                reference/drawing.html#module-networkx.drawing.layout>`__

        Returns:
             dict[str, list]: lists of x and y coordinates for the beginning and end of each edge.
             ``None`` is placed as a separator between pairs of nodes/edges.
        """
    e_x = []
    e_y = []

    for e in g.edges():

        start_x, start_y = l[e[0]]
        end_x, end_y = l[e[1]]

        e_x.append(start_x)
        e_x.append(end_x)

        e_y.append(start_y)
        e_y.append(end_y)

        e_x.append(None)
        e_y.append(None)

    return {"x": e_x, "y": e_y}


plotly_error = (
    "Plotly required for using this function. It can be installed using pip install "
    "plotly or visiting https://plot.ly/python/getting-started/#installation"
)

GREEN = "#3e9651"
RED = "#cc2529"
GREY = "#737373"
LIGHT_GREY = "#CDCDCD"
VERY_LIGHT_GREY = "#F2F2F2"

graph_node_colour = GREEN
graph_edge_colour = LIGHT_GREY
subgraph_node_colour = RED
subgraph_edge_colour = RED

graph_node_size = 14
subgraph_node_size = 16


def graph(g: nx.Graph, s: Optional[list] = None, plot_size: Tuple = (500, 500)):  # pragma: no cover
    """Creates a plot of the input graph.

    This function can plot the input graph only, or the graph with a specified subgraph highlighted.
    Graphs are plotted using the Kamada-Kawai layout with an aspect ratio of 1:1.

    **Example usage:**

    >>> graph = nx.complete_graph(10)
    >>> fig = plot.graph(graph, [0, 1, 2, 3])
    >>> fig.show()

    .. image:: ../../_static/complete_graph.png
       :width: 40%
       :align: center
       :target: javascript:void(0);

    Args:
        g (nx.Graph): input graph
        s (list): optional list of nodes comprising the subgraph to highlight
        plot_size (int): size of the plot in pixels, given as a pair of integers ``(x_size,
            y_size)``

    Returns:
         Figure: figure for graph and optionally highlighted subgraph
    """
    try:
        import plotly.graph_objects as go
    except ImportError:
        raise ImportError(plotly_error)

    l = nx.kamada_kawai_layout(g)

    g_nodes = go.Scatter(
        **_node_coords(g, l),
        mode="markers",
        hoverinfo="text",
        marker=dict(color=graph_node_colour, size=graph_node_size, line_width=2),
    )

    g_edges = go.Scatter(
        **_edge_coords(g, l),
        line=dict(width=1, color=graph_edge_colour),
        hoverinfo="none",
        mode="lines",
    )

    g_nodes.text = [str(i) for i in g.nodes()]

    layout = go.Layout(
        showlegend=False,
        hovermode="closest",
        xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
        yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
        margin=dict(b=0, l=0, r=0, t=25),
        height=plot_size[1],
        width=plot_size[0],
        plot_bgcolor="#ffffff",
    )

    if s is not None:
        s = g.subgraph(s)

        s_edges = go.Scatter(
            **_edge_coords(s, l),
            line=dict(width=2, color=subgraph_edge_colour),
            hoverinfo="none",
            mode="lines",
        )

        s_nodes = go.Scatter(
            **_node_coords(s, l),
            mode="markers",
            hoverinfo="text",
            marker=dict(color=subgraph_node_colour, size=subgraph_node_size, line_width=2),
        )

        s_nodes.text = [str(i) for i in s.nodes()]

        f = go.Figure(data=[g_edges, s_edges, g_nodes, s_nodes], layout=layout)

    else:
        f = go.Figure(data=[g_edges, g_nodes], layout=layout)

    return f


def subgraph(s: nx.Graph, plot_size: Tuple = (500, 500)):  # pragma: no cover
    """Creates a plot of the input subgraph.

    Subgraphs are plotted using the Kamada-Kawai layout with an aspect ratio of 1:1.

    **Example usage:**

    >>> graph = nx.complete_graph(10)
    >>> subgraph = graph.subgraph([0, 1, 2, 3])
    >>> fig = plot.subgraph(subgraph)
    >>> fig.show()

    .. image:: ../../_static/complete_subgraph.png
       :width: 40%
       :align: center
       :target: javascript:void(0);

    Args:
        s (nx.Graph): input subgraph
        plot_size (int): size of the plot in pixels, given as a pair of integers ``(x_size,
            y_size)``

    Returns:
         Figure: figure for subgraph
    """
    try:
        import plotly.graph_objects as go
    except ImportError:
        raise ImportError(plotly_error)

    l = nx.kamada_kawai_layout(s)

    g_edges = go.Scatter(
        **_edge_coords(s, l),
        line=dict(width=1.5, color=subgraph_edge_colour),
        hoverinfo="none",
        mode="lines",
    )

    g_nodes = go.Scatter(
        **_node_coords(s, l),
        mode="markers",
        hoverinfo="text",
        marker=dict(color=subgraph_node_colour, size=graph_node_size, line_width=2),
    )

    g_nodes.text = [str(i) for i in s.nodes()]

    layout = go.Layout(
        showlegend=False,
        hovermode="closest",
        xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
        yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
        margin=dict(b=0, l=0, r=0, t=25),
        height=plot_size[1],
        width=plot_size[0],
        plot_bgcolor="#ffffff",
    )

    f = go.Figure(data=[g_edges, g_nodes], layout=layout)

    return f


def points(
    R: np.ndarray,
    sample: Optional[list] = None,
    plot_size: Tuple = (500, 500),
    point_size: float = 30,
):  # pragma: no cover
    """Creates a plot of two-dimensional points given their input coordinates. Sampled
    points can be optionally highlighted among all points.

    **Example usage:**

    >>> R = np.random.normal(0, 1, (50, 2))
    >>> sample = [1] * 10 + [0] * 40  # select first ten points
    >>> plot.points(R, sample).show()

    .. image:: ../../_static/normal_pp.png
       :width: 40%
       :align: center
       :target: javascript:void(0);

    Args:
        R (np.array): Coordinate matrix. Rows of this array are the coordinates of the points.
        sample (list[int]): optional subset of sampled points to be highlighted
        plot_size (int): size of the plot in pixels, given as a pair of integers ``(x_size,
            y_size)``
        point_size (int): size of the points, proportional to its radius

    Returns:
         Figure: figure of points with optionally highlighted sample
    """
    try:
        import plotly.graph_objects as go
    except ImportError:
        raise ImportError(plotly_error)

    layout = go.Layout(
        showlegend=False,
        hovermode="closest",
        xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
        yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
        margin=dict(b=0, l=0, r=0, t=25),
        height=plot_size[1],
        width=plot_size[0],
        plot_bgcolor="white",
    )

    p = go.Scatter(
        x=R[:, 0],
        y=R[:, 1],
        mode="markers",
        hoverinfo="text",
        marker=dict(
            color=VERY_LIGHT_GREY, size=point_size, line=dict(color="black", width=point_size / 20)
        ),
    )

    p.text = [str(i) for i in range(len(R))]

    if sample:
        s_x = []
        s_y = []
        sampled_points = [i for i in range(len(sample)) if sample[i] > 0]
        for i in sampled_points:
            s_x.append(R[i, 0])
            s_y.append(R[i, 1])

        samp = go.Scatter(
            x=s_x,
            y=s_y,
            mode="markers",
            hoverinfo="text",
            marker=dict(
                color=RED, size=point_size, line=dict(color="black", width=point_size / 20)
            ),
        )

        samp.text = [str(i) for i in sampled_points]

        f = go.Figure(data=[p, samp], layout=layout)

    else:
        f = go.Figure(data=[p], layout=layout)

    return f


def spectrum(
    energies: list, gamma: float = 100.0, xmin: float = None, xmax: float = None
):  # pragma: no cover
    """Plots a vibronic spectrum based on input sampled energies.

    **Example usage:**

    >>> formic = data.Formic()
    >>> e = vibronic.energies(formic, formic.w, formic.wp)
    >>> full_spectrum = plot.spectrum(e, xmin=-1000, xmax=8000)
    >>> full_spectrum.show()

    .. image:: ../../_static/formic_spectrum.png
       :width: 50%
       :align: center
       :target: javascript:void(0);

    Args:
        energies (list[float]): a list of sampled energies
        gamma (float): parameter specifying the width of the Lorentzian function
        xmin (float): minimum limit of the x axis
        xmax (float): maximum limit of the x axis

    Returns:
         Figure: spectrum in the form of a histogram of energies with a Lorentzian-like curve
    """
    if len(energies) < 2:
        raise ValueError("Number of sampled energies must be at least two")
    try:
        import plotly.graph_objects as go
    except ImportError:
        raise ImportError(plotly_error)

    emin = min(energies)
    emax = max(energies)
    if xmin is None:
        xmin = emin - 0.1 * (emax - emin)
    if xmax is None:
        xmax = emax + 0.1 * (emax - emin)
    bins = int(emax - emin) // 5
    bar_width = (xmax - xmin) * 0.005
    line_width = 3.0

    h = np.histogram(energies, bins)
    X = np.linspace(xmin, xmax, int(xmax - xmin))
    L = 0
    for e in energies:
        L += (gamma / 2) ** 2 / ((X - e) ** 2 + (gamma / 2) ** 2)

    text_font = dict(color="black", family="Computer Modern")

    axis_style = dict(
        titlefont_size=30,
        tickfont=text_font,
        tickfont_size=20,
        showline=True,
        linecolor="black",
        mirror=True,
    )

    layout = go.Layout(
        yaxis=dict(title={"text": "Counts", "font": text_font}, **axis_style, rangemode="tozero"),
        xaxis=dict(
            title={"text": "Energy (cm<sup>-1</sup>)", "font": text_font},
            **axis_style,
            range=[xmin, xmax],
        ),
        plot_bgcolor="white",
        margin=dict(t=25),
        bargap=0.04,
        showlegend=False,
    )

    bars = go.Bar(x=h[1].tolist(), y=h[0].tolist(), width=bar_width, marker=dict(color=GREY))

    line = go.Scatter(x=X, y=L, mode="lines", line=dict(color=GREEN, width=line_width))

    f = go.Figure([bars, line], layout=layout)

    return f



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module implements the :class:`.Program` class which acts as a representation for quantum circuits.

Quantum circuit representation
------------------------------

The :class:`.Command` instances in the circuit form a
`strict partially ordered set <http://en.wikipedia.org/wiki/Partially_ordered_set#Strict_and_non-strict_partial_orders>`_
in the sense that the order in which the operations have to be executed is usually not completely fixed.
For example, operations acting on different subsystems always commute with each other.
We denote :math:`a < b` if :math:`a` has to be executed before :math:`b`.
Each strict partial order corresponds to a
`directed acyclic graph <http://en.wikipedia.org/wiki/Directed_acyclic_graph>`_ (DAG),
and the transitive closure of any DAG is a strict partial order.
Three different (but equivalent) representations of the circuit are used.

* Initially, the circuit is represented as a Command queue (list), listing the Commands in
  the temporal order they are applied.
* The second representation, grid, essentially mimics a quantum circuit diagram.
  It is a mapping from subsystem indices to lists of Commands touching that subsystem,
  where each list is temporally ordered.
* Finally, the quantum circuit can be represented using a DAG by making each Command a node,
  and drawing an edge from each Command to all its immediate followers along each wire it touches.
  It can be converted back into a command queue by popping a maximal element until the graph
  is empty, that is, consuming it in a topological order.
  Note that a topological order is not always unique, there may be several equivalent topological orders.

.. currentmodule:: strawberryfields.program_utils

The three representations can be converted to each other
using the functions :func:`list_to_grid`, :func:`grid_to_DAG` and :func:`DAG_to_list`.

.. currentmodule:: strawberryfields.program
"""
# pylint: disable=too-many-instance-attributes,attribute-defined-outside-init

import copy
import numbers
import warnings

import networkx as nx

import strawberryfields.circuitdrawer as sfcd
import strawberryfields.circuitspecs as specs
import strawberryfields.program_utils as pu
from .program_utils import Command, RegRef, CircuitError, RegRefError
from .parameters import FreeParameter, ParameterError


# for automodapi, do not include the classes that should appear under the top-level strawberryfields namespace
__all__ = []


class Program:
    """Represents a photonic quantum circuit.

    The program class provides a context manager for:

    * accessing the quantum register associated with the program, and
    * appending :doc:`/introduction/ops` to the program.

    Within the context, operations are appended to the program using the
    Python-embedded Blackbird syntax

    .. code-block:: python3

        ops.GateName(arg1, arg2, ...) | (q[i], q[j], ...)

    where ``ops.GateName`` is a valid quantum operation, and ``q`` is a list
    of the programs quantum modes.
    All operations are appended to the program in the order they are
    listed within the context.

    In addition, some 'meta-operations' (such as :func:`~.New` and :attr:`~.Del`)
    are provided to modify the programs quantum register itself by adding
    and deleting subsystems.

    .. note::

        Two programs can be run successively on the same engine if and only if
        the number of registers at the end of the first program matches the
        number of modes at the beginning of the second program.

        This can be enforced by constructing the second program as an explicit
        successor of the first, in which case the registers are directly copied over.

        When a Program is run or it obtains a successor, it is locked and no more
        operations can be appended to it.

    **Example:**

    .. code-block:: python3

        import strawberryfields as sf
        from strawberryfields import ops

        # create a 3 mode quantum program
        prog = sf.Program(3)

        with prog.context as q:
            ops.Sgate(0.54) | q[0]
            ops.Sgate(0.54) | q[1]
            ops.Sgate(0.54) | q[2]
            ops.BSgate(0.43, 0.1) | (q[0], q[2])
            ops.BSgate(0.43, 0.1) | (q[1], q[2])
            ops.MeasureFock() | q

    The currently active register references can be accessed using the :meth:`~Program.register` method.

    Args:
        num_subsystems (int, Program): Initial number of modes (subsystems) in the quantum register.
            Alternatively, another Program instance from which to inherit the register state.
        name (str): program name (optional)
    """

    def __init__(self, num_subsystems, name=None):
        #: str: program name
        self.name = name
        #: list[Command]: Commands constituting the quantum circuit in temporal order
        self.circuit = []
        #: bool: if True, no more Commands can be appended to the Program
        self.locked = False
        #: str, None: for compiled Programs, the short name of the target CircuitSpecs template, otherwise None
        self._target = None
        #: Program, None: for compiled Programs, this is the original, otherwise None
        self.source = None
        #: dict[str, Parameter]: free circuit parameters owned by this Program
        self.free_params = {}
        self.run_options = {}
        """dict[str, Any]: dictionary of default run options, to be passed to the engine upon
        execution of the program. Note that if the ``run_options`` dictionary is passed
        directly to :meth:`~.Engine.run`, it takes precedence over the run options specified
        here.
        """

        # create subsystem references
        # Program keeps track of the state of the quantum register using a dictionary of :class:`RegRef` objects.
        if isinstance(num_subsystems, numbers.Integral):
            #: int: initial number of subsystems
            self.init_num_subsystems = num_subsystems
            #: dict[int, RegRef]: mapping from subsystem indices to corresponding RegRef objects
            self.reg_refs = {}
            #: set[int]: created subsystem indices that have not been used (operated on) yet
            self.unused_indices = set()
            self._add_subsystems(num_subsystems)
        elif isinstance(num_subsystems, Program):
            # it's the parent program
            parent = num_subsystems
            # copy the RegRef state from the parent program
            parent.lock()  # make sure the parent isn't accidentally updated by the user
            self.init_num_subsystems = parent.num_subsystems
            self.reg_refs = copy.deepcopy(parent.reg_refs)  # independent copy of the RegRefs
            self.unused_indices = copy.copy(parent.unused_indices)
        else:
            raise TypeError(
                "First argument must be either the number of subsystems or the parent Program."
            )

        # save the initial regref state
        #: dict[int, RegRef]: like reg_refs
        self.init_reg_refs = copy.deepcopy(self.reg_refs)
        #: set[int]: like unused_indices
        self.init_unused_indices = copy.copy(self.unused_indices)

    def __str__(self):
        """String representation."""
        return self.__class__.__name__ + "({}, {}->{} subsystems, compiled for '{}')".format(
            self.name, self.init_num_subsystems, self.num_subsystems, self.target
        )

    def __len__(self):
        """Program length.

        Returns:
            int: number of Commands in the program
        """
        return len(self.circuit)

    def print(self, print_fn=print):
        """Print the program contents using Blackbird syntax.

        **Example:**

        .. code-block:: python

            # create a 3 mode quantum program
            prog = sf.Program(3)

            with prog.context as q:
                ops.Sgate(0.54) | q[0]
                ops.Sgate(0.54) | q[1]
                ops.Sgate(0.54) | q[2]
                ops.BSgate(0.43, 0.1) | (q[0], q[2])
                ops.BSgate(0.43, 0.1) | (q[1], q[2])
                ops.MeasureFock() | q

        >>> prog.print()
        Sgate(0.54, 0) | (q[0])
        Sgate(0.54, 0) | (q[1])
        Sgate(0.54, 0) | (q[2])
        BSgate(0.43, 0.1) | (q[0], q[2])
        BSgate(0.43, 0.1) | (q[1], q[2])
        MeasureFock | (q[0], q[1], q[2])

        Args:
            print_fn (function): optional custom function to use for string printing
        """
        for k in self.circuit:
            print_fn(k)

    @property
    def context(self):
        """Syntactic sugar for defining a Program using the :code:`with` statement.

        The Program object itself acts as the context manager.
        """
        return self

    def __enter__(self):
        """Enter the context for this program.

        Returns:
            tuple[RegRef]: subsystem references
        """
        if pu.Program_current_context is None:
            pu.Program_current_context = self
        else:
            raise RuntimeError("Only one Program context can be active at a time.")
        return self.register

    def __exit__(self, ex_type, ex_value, ex_tb):
        """Exit the quantum circuit context."""
        pu.Program_current_context = None

    # =================================================
    #  RegRef accounting
    # =================================================
    @property
    def register(self):
        """Return a tuple of all the currently valid quantum modes.

        Returns:
            tuple[RegRef]: valid subsystem references
        """
        return tuple(r for r in self.reg_refs.values() if r.active)

    @property
    def num_subsystems(self):
        """Return the current number of valid quantum modes.

        Returns:
            int: number of currently valid register subsystems
        """
        return len(self.register)

    def _clear_regrefs(self):
        """Clear any measurement values stored in the RegRefs.

        Called by :class:`~.engine.Engine` when resetting the backend.
        """
        for r in self.reg_refs.values():
            r.val = None

    def _add_subsystems(self, n):
        """Create new subsystem references, add them to the reg_ref dictionary.

        To avoid discrepancies with the backend this method must not be called directly,
        but rather indirectly by using :func:`~strawberryfields.ops.New`
        in a Program context.

        .. note:: This is the only place where :class:`RegRef` instances are constructed.

        Args:
            n (int): number of subsystems to add
        Returns:
            tuple[RegRef]: tuple of the newly added subsystem references
        """
        if self.locked:
            raise CircuitError("The Program is locked, no new subsystems can be created.")
        if not isinstance(n, numbers.Integral) or n < 1:
            raise ValueError("Number of added subsystems {} is not a positive integer.".format(n))

        first_unassigned_index = len(self.reg_refs)
        # create a list of RegRefs
        inds = [first_unassigned_index + i for i in range(n)]
        refs = tuple(RegRef(i) for i in inds)
        # add them to the index map
        for r in refs:
            self.reg_refs[r.ind] = r
        # all the newly reserved indices are unused for now
        self.unused_indices.update(inds)
        return refs

    def _delete_subsystems(self, refs):
        """Delete existing subsystem references.

        To avoid discrepancies with the backend this method must not be called directly,
        but rather indirectly by using :class:`~strawberryfields.ops._Delete` instances
        in the Program context.

        Args:
          refs (Sequence[RegRef]): subsystems to delete
        """
        # NOTE: refs have already been through _test_regrefs() in append() and thus should be valid
        for r in refs:
            # mark the RegRef as deleted
            r.active = False
            # self.reg_refs[r.ind].active = False
        # NOTE: deleted indices are *not* removed from self.unused_indices

    def lock(self):
        """Finalize the program.

        When a Program is locked, no more Commands can be appended to it.
        The locking happens when the program is run, compiled, or a successor Program is constructed,
        in order to ensure that the RegRef state of the Program does not change anymore.
        """
        self.locked = True

    def can_follow(self, prev):
        """Check whether this program can follow the given program.

        This requires that the final RegRef state of the first program matches
        the initial RegRef state of the second program, i.e., they have the same number
        number of RegRefs, all with identical indices and activity states.

        Args:
            prev (Program): preceding program fragment
        Returns:
            bool: True if the Program can follow prev
        """
        # TODO NOTE unused_indices is not compared here, in order to allow program fragment repetition
        return self.init_reg_refs == prev.reg_refs

    def _index_to_regref(self, ind):
        """Try to find a RegRef corresponding to a given subsystem index.

        Args:
            ind (int): subsystem index
        Returns:
            RegRef: corresponding register reference
        Raises:
            .RegRefError: if the subsystem cannot be found, or is invalid
        """
        # index must be found in the dict
        if ind not in self.reg_refs:
            raise RegRefError("Subsystem {} does not exist.".format(ind))
        return self.reg_refs[ind]

    def _test_regrefs(self, reg):
        """Make sure reg is a valid selection of subsystems, convert them to RegRefs.

        A register reference is valid if it is properly recorded in self.reg_refs
        and has not been deleted. The selection is valid if it contains only
        valid RegRefs and no subsystem is repeated.

        Args:
            reg (Iterable[int, RegRef]): subsystem references
        Returns:
            list[RegRef]: converted subsystem references
        Raises:
            .RegRefError: if an invalid subsystem reference is found
        """
        temp = []
        for rr in reg:
            # must be either an integer or a RegRef
            if isinstance(rr, RegRef):
                # regref must be found in the dict values (the RegRefs are compared using __eq__, which, since we do not define it, defaults to "is")
                if rr not in self.reg_refs.values():
                    raise RegRefError("Unknown RegRef.")
                if self.reg_refs[rr.ind] is not rr:
                    raise RegRefError("RegRef state has become inconsistent.")
            elif isinstance(rr, numbers.Integral):
                rr = self._index_to_regref(rr)
            else:
                raise RegRefError("Subsystems can only be indexed using integers and RegRefs.")

            if not rr.active:
                raise RegRefError("Subsystem {} has already been deleted.".format(rr.ind))
            if rr in temp:
                raise RegRefError("Trying to act on the same subsystem more than once.")
            temp.append(rr)
        return temp

    def append(self, op, reg):
        """Append a command to the program.

        Args:
            op (Operation): quantum operation
            reg (list[int, RegRef]): register subsystem(s) to apply it to
        Returns:
            list[RegRef]: subsystem list as RegRefs
        """
        if self.locked:
            raise CircuitError("The Program is locked, no more Commands can be appended to it.")

        # test that the target subsystem references are ok
        reg = self._test_regrefs(reg)
        # also test possible Parameter-related dependencies
        self._test_regrefs(op.measurement_deps)
        for rr in reg:
            # it's used now
            self.unused_indices.discard(rr.ind)
        self.circuit.append(Command(op, reg))
        return reg

    def _linked_copy(self):
        """Create a copy of the Program, linked to the original.

        Both the original and the copy are :meth:`locked <lock>`, since they share their RegRefs.
        FreeParameters are also shared.

        Returns:
            Program: a copy of the Program
        """
        self.lock()
        p = copy.copy(self)  # shares RegRefs with the source
        # link to the original source Program
        if self.source is None:
            p.source = self
        else:
            p.source = self.source
        return p

    def compile(self, target, **kwargs):
        """Compile the program targeting the given circuit specification.

        Validates the program against the given target, making sure all the
        :doc:`/introduction/ops` used are accepted by the target specification.

        Additionally, depending on the target, the compilation may modify the quantum circuit
        into an equivalent circuit, e.g., by decomposing certain gates into sequences
        of simpler gates, or optimizing the gate ordering using commutation rules.

        **Example:**

        The ``gbs`` compile target will
        compile a circuit consisting of Gaussian operations and Fock measurements
        into canonical Gaussian boson sampling form.

        >>> prog2 = prog.compile('gbs')

        Args:
            target (str, ~strawberryfields.circuitspecs.CircuitSpecs): short name of the target
                circuit specification, or the specification object itself

        Keyword Args:
            optimize (bool): If True, try to optimize the program by merging and canceling gates.
                The default is False.
            warn_connected (bool): If True, the user is warned if the quantum circuit is not weakly
                connected. The default is True.

        Returns:
            Program: compiled program
        """
        if isinstance(target, specs.CircuitSpecs):
            db = target
            target = db.short_name
        elif target in specs.circuit_db:
            db = specs.circuit_db[target]()
        else:
            raise ValueError(
                "Could not find target '{}' in the Strawberry Fields circuit database.".format(
                    target
                )
            )

        if db.modes is not None:
            # subsystems may be created and destroyed, this is total number that has ever existed
            modes_total = len(self.reg_refs)
            if modes_total > db.modes:
                raise CircuitError(
                    "This program requires {} modes, but the target '{}' "
                    "only supports a {}-mode program".format(modes_total, target, db.modes)
                )

        seq = db.decompose(self.circuit)

        if kwargs.get("warn_connected", True):
            DAG = pu.list_to_DAG(seq)
            temp = nx.algorithms.components.number_weakly_connected_components(DAG)
            if temp > 1:
                warnings.warn("The circuit consists of {} disconnected components.".format(temp))

        # run optimizations
        if kwargs.get("optimize", False):
            seq = pu.optimize_circuit(seq)

        # does the circuit spec  have its own compilation method?
        if db.compile is not None:
            seq = db.compile(seq, self.register)

        # create the compiled Program
        compiled = self._linked_copy()
        compiled.circuit = seq
        compiled._target = db.short_name

        # get run options of compiled program
        # for the moment, shots is the only supported run option.
        if "shots" in kwargs:
            compiled.run_options["shots"] = kwargs["shots"]

        compiled.backend_options = {}
        if "cutoff_dim" in kwargs:
            compiled.backend_options["cutoff_dim"] = kwargs["cutoff_dim"]

        return compiled

    def optimize(self):
        """Simplify and optimize the program.

        The simplifications are based on the algebraic properties of the gates,
        e.g., combining two consecutive gates of the same gate family.

        Returns a copy of the program, sharing RegRefs with the original.

        See :func:`~strawberryfields.program_utils.optimize_circuit`.

        Returns:
            Program: optimized copy of the program
        """
        opt = self._linked_copy()
        opt.circuit = pu.optimize_circuit(self.circuit)
        return opt

    def draw_circuit(self, tex_dir="./circuit_tex", write_to_file=True):
        r"""Draw the circuit using the Qcircuit :math:`\LaTeX` package.

        This will generate the LaTeX code required to draw the quantum circuit
        diagram corresponding to the Program.


        The drawing of the following Xanadu supported operations are currently supported:

        .. rst-class:: docstable

        +-------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |     Gate type     |                                                                            Supported gates                                                                             |
        +===================+========================================================================================================================================================================+
        | Single mode gates | :class:`~.Dgate`, :class:`~.Xgate`, :class:`~.Zgate`, :class:`~.Sgate`, :class:`~.Rgate`, :class:`~.Pgate`, :class:`~.Vgate`, :class:`~.Kgate`, :class:`~.Fouriergate` |
        +-------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        | Two mode gates    | :class:`~.BSgate`, :class:`~.S2gate`, :class:`~.CXgate`, :class:`~.CZgate`, :class:`~.CKgate`                                                                          |
        +-------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------+

        .. note::

            Measurement operations :class:`~.MeasureHomodyne`, :class:`~.MeasureHeterodyne`,
            and :class:`~.MeasureFock` are not currently supported.

        Args:
            tex_dir (str): relative directory for latex document output
            write_to_file (bool): if False, no output file is created

        Returns:
            list[str]: filename of the written tex document and the written tex content
        """
        drawer = sfcd.Circuit(wires=self.init_num_subsystems)
        self.print(drawer.parse_op)
        tex = drawer.dump_to_document()

        document = None
        if write_to_file:
            document = drawer.compile_document(tex_dir=tex_dir)

        return [document, tex]

    @property
    def target(self):
        """The target specification the program has been compiled against.

        If the program has not been compiled, this will return ``None``.

        Returns:
            str or None: the short name of the target CircuitSpecs template if
            compiled, otherwise None
        """
        return self._target

    def params(self, *args):
        """Create and access free circuit parameters.

        Returns the named free parameters. If a parameter does not exist yet, it is created and returned.

        Args:
            *args (tuple[str]): name(s) of the free parameters to access

        Returns:
            FreeParameter, list[FreeParameter]: requested parameter(s)
        """
        ret = []
        for a in args:
            if not isinstance(a, str):
                raise TypeError("Parameter names must be strings.")

            if a not in self.free_params:
                if self.locked:
                    raise CircuitError(
                        "The Program is locked, no more free parameters can be created."
                    )
                p = FreeParameter(a)
                self.free_params[a] = p
            else:
                p = self.free_params[a]
            ret.append(p)

        if len(ret) == 1:
            return ret[0]
        return ret

    def bind_params(self, binding):
        """Binds the free parameters of the program to the given values.

        Args:
            binding (dict[Union[str, FreeParameter], Any]): mapping from parameter names (or the
                parameters themselves) to parameter values

        Raises:
            ParameterError: tried to bind an unknown parameter
        """
        for k, v in binding.items():
            temp = self.free_params.get(k)  # it's a name
            if temp:
                temp.val = v
            elif k in self.free_params.values():  # it's a parameter
                k.val = v
            else:
                raise ParameterError("Unknown free parameter '{}'".format(k))



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Functions used for computing molecular vibronic spectra with GBS.

In vibronic spectroscopy, incoming light simultaneously excites a molecule to higher electronic and
vibrational states. The difference in the energies of the initial and final states
determines the frequency of the photons that are absorbed by the molecule. The probability of
the transition between two specific vibrational states are referred to as Franck-Condon factors
(FCFs). They determine the intensity of absorption peaks in a vibronic spectrum. A GBS
device can be programmed to provide samples that can be processed to obtain the molecular
vibronic spectra. Theoretical background on computing vibronic spectra using GBS can be found in
:cite:`huh2015boson` and :cite:`quesada2019franck`.

.. seealso::

    :ref:`apps-vibronic-tutorial`

GBS parameters
--------------

The Franck-Condon factor is given by

.. math::
    FCF = \left | \left \langle \mathbf{m} |U_{Dok}| \mathbf{n} \right \rangle \right | ^ 2,

where :math:`|\mathbf{m}\rangle` and :math:`|\mathbf{n}\rangle` are the final and initial states,
respectively, and :math:`U_{Dok}` is known as the Doktorov operator which can be written in terms
of displacement :math:`D_\alpha`, squeezing :math:`\Sigma_{r}`, and interferometer :math:`U_{1},
U_{2}` operators as

.. math::
    U_{Dok} = D_\alpha U_{2} \Sigma_{r} U_{1}.

A GBS device can be programmed with the :math:`U_{Dok}` operator obtained for a given molecule. The
function :func:`gbs_params` transforms molecular parameters, namely vibrational frequencies,
displacement vector and Duschinsky matrix, to the required GBS parameters. Additionally, this
function computes two-mode squeezing parameters :math:`t`, from the molecule's temperature, which
are required by the GBS algorithm to compute vibronic spectra for molecules at finite temperature.

Energies from samples
---------------------
In the GBS algorithm for vibronic spectra, samples are identified with transition energies. The
most frequently sampled energies correspond to peaks of the vibronic spectrum. The function
:func:`energies` converts samples to energies.
"""
from typing import Tuple, Union

from scipy.constants import c, h, k

import numpy as np


def gbs_params(
    w: np.ndarray, wp: np.ndarray, Ud: np.ndarray, delta: np.ndarray, T: float = 0
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    r"""Converts molecular information to GBS gate parameters.

    **Example usage:**

    >>> formic = data.Formic()
    >>> w = formic.w  # ground state frequencies
    >>> wp = formic.wp  # excited state frequencies
    >>> Ud = formic.Ud  # Duschinsky matrix
    >>> delta = formic.delta  # displacement vector
    >>> T = 0  # temperature
    >>> p = gbs_params(w, wp, Ud, delta, T)

    Args:
        w (array): normal mode frequencies of the initial electronic state in units of
            :math:`\mbox{cm}^{-1}`
        wp (array): normal mode frequencies of the final electronic state in units of
            :math:`\mbox{cm}^{-1}`
        Ud (array): Duschinsky matrix
        delta (array): Displacement vector, with entries :math:`\delta_i=\sqrt{\omega'_i/\hbar}d_i`,
            and :math:`d_i` is the Duschinsky displacement
        T (float): temperature (Kelvin)

    Returns:
        tuple[array, array, array, array, array]: the two-mode squeezing parameters :math:`t`,
        the first interferometer unitary matrix :math:`U_{1}`, the squeezing parameters :math:`r`,
        the second interferometer unitary matrix :math:`U_{2}`, and the displacement
        parameters :math:`\alpha`
    """
    if T < 0:
        raise ValueError("Temperature must be zero or positive")
    if T > 0:
        t = np.arctanh(np.exp(-0.5 * h * (w * c * 100) / (k * T)))
    else:
        t = np.zeros(len(w))

    U2, s, U1 = np.linalg.svd(np.diag(wp ** 0.5) @ Ud @ np.diag(w ** -0.5))
    alpha = delta / np.sqrt(2)

    return t, U1, np.log(s), U2, alpha


def energies(samples: list, w: np.ndarray, wp: np.ndarray) -> Union[list, float]:
    r"""Computes the energy :math:`E = \sum_{k=1}^{N}m_k\omega'_k - \sum_{k=N+1}^{2N}n_k\omega_k`
    of each GBS sample in units of :math:`\text{cm}^{-1}`.

    **Example usage:**

    >>> samples = [[1, 1, 0, 0, 0, 0], [1, 2, 0, 0, 1, 1]]
    >>> w  = np.array([300.0, 200.0, 100.0])
    >>> wp = np.array([700.0, 600.0, 500.0])
    >>> energies(samples, w, wp)
    [1300.0, 1600.0]

    Args:
        samples (list[list[int]] or list[int]): a list of samples from GBS, or alternatively a
            single sample
        w (array): normal mode frequencies of initial state in units of :math:`\text{cm}^{-1}`
        wp (array): normal mode frequencies of final state in units of :math:`\text{cm}^{-1}`

    Returns:
        list[float] or float: list of GBS sample energies in units of :math:`\text{cm}^{-1}`, or
        a single sample energy if only one sample is input
    """
    if not isinstance(samples[0], list):
        return np.dot(samples[: len(samples) // 2], wp) - np.dot(samples[len(samples) // 2 :], w)

    return [np.dot(s[: len(s) // 2], wp) - np.dot(s[len(s) // 2 :], w) for s in samples]



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Miscellaneous Fock backend operations"""
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals


import functools
import string
from itertools import product

import numpy as np
from numpy import sqrt, sinh, cosh, tanh, array, exp
from numpy.polynomial.hermite import hermval as H

from scipy.special import factorial as fac
from scipy.linalg import expm as matrixExp

from thewalrus.fock_gradients import Dgate, Sgate, S2gate, BSgate


def_type = np.complex128
indices = string.ascii_lowercase


def genOfRange(size):
    """
    Converts a range into a generator.
    """
    for i in range(size):
        yield i


def genOfTuple(t):
    """
    Converts a tuple into a generator
    """
    for val in t:
        yield val


def indexRange(lst, trunc):
    """
    Returns a generator ranging over the possible values for unspecified
    indices in `lst`.

    Example:
        .. code-block:: python

                >>> for i in indexRange([0, None, 1, None], 3): print i
                (0, 0, 1, 0)
                (0, 0, 1, 1)
                (0, 0, 1, 2)
                (0, 1, 1, 0)
                (0, 1, 1, 1)
                (0, 1, 1, 2)
                (0, 2, 1, 0)
                (0, 2, 1, 1)
                (0, 2, 1, 2)

    Args:
        lst (list<int or None>): a list of (possible unspecified) integers
        trunc (int): the number to range unspecified values up to

    Returns:
        Generator of ints
    """

    for vals in product(*([range(trunc) for x in lst if x is None])):
        gen = genOfTuple(vals)
        yield [next(gen) if v is None else v for v in lst]  # pylint: disable=stop-iteration-return


def index(lst, trunc):
    """
    Converts an n-ary index to a 1-dimensional index.
    """
    return sum([lst[i] * trunc ** (len(lst) - i - 1) for i in range(len(lst))])


def unIndex(i, n, trunc):
    """
    Converts a 1-dimensional index ``i`` with truncation ``trunc`` and
    number of modes ``n`` to a n-ary index.
    """
    return [i // trunc ** (n - 1 - m) % trunc for m in range(n)]


def sliceExp(axes, ind, n):
    """
    Generates a slice expression for a list of pairs of axes (modes) and indices.
    """
    return [ind[i] if i in axes else slice(None, None, None) for i in range(n)]


def abssqr(z):
    r"""
    Given :math:`z` returns :math:`|z|^2`.
    """
    return z.real ** 2 + z.imag ** 2


def dagger(mat):
    r"""
    Given :math:`U` returns :math:`U^\dagger`.
    """
    return mat.conj().T


def mix(state, n):
    """
    Transforms a pure state into a mixed state. Does not do any checks on the
    shape of the input state.
    """

    left_str = [indices[i] for i in range(0, 2 * n, 2)]
    right_str = [indices[i] for i in range(1, 2 * n, 2)]
    out_str = [indices[: 2 * n]]
    einstr = "".join(left_str + [","] + right_str + ["->"] + out_str)
    return np.einsum(einstr, state, state.conj())


def diagonal(state, n):
    """
    Computes the diagonal of a density matrix.
    """

    left_str = [indices[i] + indices[i] for i in range(n)]
    out_str = [indices[:n]]
    einstr = "".join(left_str + ["->"] + out_str)
    return np.einsum(einstr, state)


def trace(state, n):
    """
    Computes the trace of a density matrix.
    """

    left_str = [indices[i] + indices[i] for i in range(n)]
    einstr = "".join(left_str)
    return np.einsum(einstr, state)


def partial_trace(state, n, modes):
    """
    Computes the partial trace of a state over the modes in `modes`.

    Expects state to be in mixed state form.
    """
    left_str = [
        indices[2 * i] + indices[2 * i] if i in modes else indices[2 * i : 2 * i + 2]
        for i in range(n)
    ]
    out_str = ["" if i in modes else indices[2 * i : 2 * i + 2] for i in range(n)]
    einstr = "".join(left_str + ["->"] + out_str)

    return np.einsum(einstr, state)


def tensor(u, v, n, pure, pos=None):
    """
    Returns the tensor product of `u` and `v`, optionally spliced into a
    at location `pos`.
    """

    w = np.tensordot(u, v, axes=0)

    if pos is not None:
        if pure:
            scale = 1
        else:
            scale = 2
        for i in range(v.ndim):
            w = np.rollaxis(w, scale * n + i, scale * pos + i)

    return w


def project_reset(modes, x, state, pure, n, trunc):
    r"""
    Applies the operator :math:`\ket{00\dots 0}\bra{\mathbf{x}}` to the
    modes in `modes`.
    """
    inSlice = tuple(sliceExp(modes, dict(zip(modes, x)), n))
    outSlice = tuple(sliceExp(modes, dict(zip(modes, [0] * len(modes))), n))

    def intersperse(lst):
        # pylint: disable=missing-docstring
        return tuple([lst[i // 2] for i in range(len(lst) * 2)])

    if pure:
        ret = np.zeros([trunc for i in range(n)], dtype=def_type)
        ret[outSlice] = state[inSlice]
    else:
        ret = np.zeros([trunc for i in range(n * 2)], dtype=def_type)
        ret[intersperse(outSlice)] = state[intersperse(inSlice)]

    return ret


# ============================================
#
# Gates
#
# ============================================


@functools.lru_cache()
def a(trunc):
    r"""
    The annihilation operator :math:`a`.
    """
    ret = np.zeros((trunc, trunc), dtype=def_type)
    for i in range(1, trunc):
        ret[i - 1][i] = sqrt(i)
    return ret


@functools.lru_cache()
def displacement(alpha, trunc):
    r"""The displacement operator :math:`D(\alpha)`.

    Uses the `Dgate operation from The Walrus`_ to calculate the displacement.

    .. _`Dgate operation from The Walrus`: https://the-walrus.readthedocs.io/en/latest/code/api/thewalrus.fock_gradients.Dgate.html

    Args:
            alpha (complex): the displacement
            trunc (int): the Fock cutoff
    """

    r = np.abs(alpha)
    theta = np.angle(alpha)

    ret, _, _ = Dgate(r, theta, cutoff=trunc)

    return ret


@functools.lru_cache()
def squeezing(r, theta, trunc):
    r"""The squeezing operator :math:`S(re^{i\theta})`.

    Uses the `Sgate operation from The Walrus`_ to calculate the squeezing.

    .. _`Sgate operation from The Walrus`: https://the-walrus.readthedocs.io/en/latest/code/api/thewalrus.fock_gradients.Sgate.html

    Args:
            r (float): the magnitude of the squeezing in the
                    x direction
            theta (float): the squeezing angle
            trunc (int): the Fock cutoff
    """

    ret, _, _ = Sgate(r, theta, cutoff=trunc)

    return ret


@functools.lru_cache()
def two_mode_squeezing(r, theta, trunc):
    r"""The two-mode squeezing operator :math:`S_2(re^{i\theta})`.

    Args:
        r (float): two-mode squeezing magnitude
        theta (float): two-mode squeezing phase
        trunc (int): Fock ladder cutoff
    """
    ret, _, _ = S2gate(r, theta, cutoff=trunc)

    ret = np.transpose(ret, [0, 2, 1, 3])

    return ret


@functools.lru_cache()
def kerr(kappa, trunc):
    r"""
    The Kerr interaction :math:`K(\kappa)`.
    """
    n = np.arange(trunc)
    ret = np.diag(np.exp(1j * kappa * n ** 2))
    return ret


@functools.lru_cache()
def cross_kerr(kappa, trunc):
    r"""
    The cross-Kerr interaction :math:`CK(\kappa)`.
    """
    n1 = np.arange(trunc)[None, :]
    n2 = np.arange(trunc)[:, None]
    n1n2 = np.ravel(n1 * n2)
    ret = np.diag(np.exp(1j * kappa * n1n2)).reshape([trunc] * 4).swapaxes(1, 2)
    return ret


@functools.lru_cache()
def cubicPhase(gamma, hbar, trunc):
    r"""
    The cubic phase gate :math:`\exp{(i\frac{\gamma}{3\hbar}\hat{x}^3)}`.
    """
    a_ = a(trunc)
    x = (a_ + np.conj(a_).T) * np.sqrt(hbar / 2)
    x3 = x @ x @ x
    ret = matrixExp(1j * gamma / (3 * hbar) * x3)

    return ret


@functools.lru_cache()
def phase(theta, trunc):
    r"""
    The phase gate :math:`R(\theta)`
    """
    return np.array(np.diag([exp(1j * n * theta) for n in range(trunc)]), dtype=def_type)


@functools.lru_cache()
def beamsplitter(t, r, phi, trunc):
    r""" The beamsplitter :math:`B(cos^{-1} t, phi)`.

    Uses the `BSgate operation from The Walrus`_ to calculate the beamsplitter.

    .. _`BSgate operation from The Walrus`: https://the-walrus.readthedocs.io/en/latest/code/api/thewalrus.fock_gradients.BSgate.html
    """
    # pylint: disable=bad-whitespace

    theta = np.arccos(t)
    BS_tw, _, _ = BSgate(theta, phi, cutoff=trunc)

    # TODO: Transpose needed because of different conventions in SF and The Walrus. Remove when The Walrus is updated.
    return BS_tw.transpose((0, 2, 1, 3))


@functools.lru_cache()
def proj(i, j, trunc):
    r"""
    The projector :math:`P = \ket{j}\bra{i}`.
    """
    P = np.zeros((trunc, trunc), dtype=def_type)
    P[j][i] = 1.0 + 0.0j
    return P


# ============================================
#
# State vectors
#
# ============================================


def vacuumState(n, trunc):
    r"""
    The `n`-mode vacuum state :math:`\ket{00\dots 0}`
    """

    state = np.zeros([trunc for i in range(n)], dtype=def_type)
    state.ravel()[0] = 1.0 + 0.0j
    return state


def vacuumStateMixed(n, trunc):
    r"""
    The `n`-mode mixed vacuum state :math:`\ket{00\dots 0}\bra{00\dots 0}`
    """

    state = np.zeros([trunc for i in range(n * 2)], dtype=def_type)
    state.ravel()[0] = 1.0 + 0.0j
    return state


@functools.lru_cache()
def fockState(n, trunc):
    r"""
    The Fock state :math:`\ket{n}`.
    """
    return array([1.0 + 0.0j if i == n else 0.0 + 0.0j for i in range(trunc)])


@functools.lru_cache()
def coherentState(alpha, trunc):
    r"""
    The coherent state :math:`D(\alpha)\ket{0}`.
    """

    def entry(n):
        """coherent summation term"""
        return alpha ** n / sqrt(fac(n))

    return exp(-abssqr(alpha) / 2) * array([entry(n) for n in range(trunc)])


@functools.lru_cache()
def squeezedState(r, theta, trunc):
    r"""
    The squeezed state :math:`S(re^{i\theta})`.
    """

    def entry(n):
        """squeezed summation term"""
        return (sqrt(fac(2 * n)) / (2 ** n * fac(n))) * (-exp(1j * theta) * tanh(r)) ** n

    vec = array([entry(n // 2) if n % 2 == 0 else 0.0 + 0.0j for n in range(trunc)])
    return sqrt(1 / cosh(r)) * vec


@functools.lru_cache()
def displacedSqueezed(alpha, r, phi, trunc):
    r"""
    The displaced squeezed state :math:`\ket{\alpha,\zeta} = D(\alpha)S(r\exp{(i\phi)})\ket{0}`.
    """
    if r == 0:
        return coherentState(alpha, trunc)

    if alpha == 0:
        return squeezedState(r, phi, trunc)

    ph = np.exp(1j * phi)
    ch = cosh(r)
    sh = sinh(r)
    th = tanh(r)

    gamma = alpha * ch + np.conj(alpha) * ph * sh
    hermite_arg = gamma / np.sqrt(ph * np.sinh(2 * r) + 1e-10)

    # normalization constant
    N = np.exp(-0.5 * np.abs(alpha) ** 2 - 0.5 * np.conj(alpha) ** 2 * ph * th)

    coeff = np.array([(0.5 * ph * th) ** (n / 2) / np.sqrt(fac(n) * ch) for n in range(trunc)])
    vec = np.array([H(hermite_arg, row) for row in np.diag(coeff)])
    state = N * vec

    return state


@functools.lru_cache()
def thermalState(nbar, trunc):
    r"""
    The thermal state :math:`\rho(\overline{nbar})`.
    """
    if nbar == 0:
        st = fockState(0, trunc)
        state = np.outer(st, st.conjugate())
    else:
        coeff = np.array([nbar ** n / (nbar + 1) ** (n + 1) for n in range(trunc)])
        state = np.diag(coeff)

    return state


# ============================================
#
# Channels (Kraus operators)
#
# ============================================


@functools.lru_cache()
def lossChannel(T, trunc):
    r"""
    The Kraus operators for the loss channel :math:`\mathcal{N}(T)`.
    """

    TToAdaggerA = np.array(np.diag([T ** (i / 2) for i in range(trunc)]), dtype=def_type)

    def aToN(n):
        """the nth matrix power of the annihilation operator matrix a"""
        return np.linalg.matrix_power(a(trunc), n)

    def E(n):
        """the loss channel amplitudes in the Fock basis"""
        return ((1 - T) / T) ** (n / 2) * np.dot(aToN(n) / sqrt(fac(n)), TToAdaggerA)

    if T == 0:
        return [proj(i, 0, trunc) for i in range(trunc)]

    return [E(n) for n in range(trunc)]


# ============================================
#
# Misc
#
# ============================================


@functools.lru_cache()
def hermiteVals(q_mag, num_bins, m_omega_over_hbar, trunc):
    """
    Helper function for homodyne measurements. Computes a range of (physicist's)
    Hermite polynomials at a particular vector.
    """
    q_tensor = np.linspace(-q_mag, q_mag, num_bins)
    x = np.sqrt(m_omega_over_hbar) * q_tensor

    Hvals = [None] * trunc
    Hvals[0] = 1
    Hvals[1] = 2 * x
    for i in range(2, trunc):
        Hvals[i] = 2 * x * Hvals[i - 1] - 2 * (i - 1) * Hvals[i - 2]

    return q_tensor, Hvals



#!/usr/bin/env python3
import strawberryfields as sf
from strawberryfields.ops import *
from numpy import pi

# initialize engine and program objects
eng = sf.Engine(backend="fock", backend_options={"cutoff_dim": 5})
ham_simulation = sf.Program(2)

# set the Hamiltonian parameters
J = 1           # hopping transition
U = 1.5         # on-site interaction
k = 20          # Lie product decomposition terms
t = 1.086       # timestep
theta = -J*t/k  
r = -U*t/(2*k)

with ham_simulation.context as q:
    # prepare the initial state
    Fock(2) | q[0]

    # Two node tight-binding
    # Hamiltonian simulation

    for i in range(k):
        BSgate(theta, pi/2) | (q[0], q[1])
        Kgate(r)  | q[0]
        Rgate(-r) | q[0]
        Kgate(r)  | q[1]
        Rgate(-r) | q[1]
    # end circuit

# run the engine
results = eng.run(ham_simulation)
# the output state probabilities
print(results.state.fock_prob([0,2]))
print(results.state.fock_prob([1,1]))
print(results.state.fock_prob([2,0]))
